####################################################################################################
# neuropythy/retinotopy/models.py
# Importing and interpreting of flat mesh models for registration.
# By Noah C. Benson

import numpy                 as     np
import numpy.linalg          as     npla
import scipy                 as     sp
import scipy.spatial         as     space
import pyrsistent            as     pyr
import os, gzip, types, six, pimms

from ..           import geometry as geo
from ..           import mri      as mri
from ..java       import (java_link, serialize_numpy,
                                     to_java_doubles, to_java_ints, to_java_array)
from ..util       import (to_affine, library_path, is_tuple, is_list)
from ..io         import importer

# These two variables are intended to provide default orderings to visual areas (but in general,
# visual areas should be referred to by name OR as a number paired with a model).
visual_area_names = (None,
                     'V1', 'V2', 'V3', 'hV4', 'VO1', 'VO2', 'LO1', 'LO2',
                     'TO1', 'TO2', 'V3b', 'V3a')
visual_area_numbers = pyr.pmap({v:k for (k,v) in enumerate(visual_area_names)})
visual_area_field_signs = (0,
                           -1, 1, -1, 1, -1, 1, 1, -1,
                           1, -1, 1, -1)
visual_area_field_signs = pyr.pmap(
    {k:v for (k,v) in zip(tuple(range(len(visual_area_numbers))) + visual_area_names,
                          visual_area_field_signs * 2)})

@pimms.immutable
class RetinotopyModel(object):
    '''
    RetinotopyModel is a class designed to be inherited by other models of retinotopy; any class
    that inherits from RetinotopyModel must implement the following methods to work properly with
    the registration system of the neuropythy library.
    '''

    def __init__(self, area_name_to_id):
        self.area_name_to_id = area_name_to_id
    @pimms.param
    def area_name_to_id(vai):
        '''
        mdl.area_name_to_id is a persistent map whose keys are area names (such as 'V1' or 'hV4')
        and whose values are the area id (a number greater than 0) for that area.
        mdl.area_name_to_id is a parameter which may be provided as a lsit of area names, in which
        case the first is assumed to be area 1, the next area 2, etc.
        '''
        if vai is None: return None
        if not pimms.is_map(vai): return pyr.pmap({nm:(ii+1) for (ii,nm) in enumerate(vai)})
        elif pimms.is_pmap(vai): return vai
        else: return pyr.pmap(vai)
    @pimms.value
    def area_id_to_name(area_name_to_id):
        '''
        mdl.area_id_to_name is a persistent map whose keys are area id's and whose values are the
        associated area's name.
        '''
        if area_name_to_id is None: return None
        return pyr.pmap({v:k for (k,v) in six.iteritems(area_name_to_id)})
    # Methods that must be overloaded!
    def angle_to_cortex(self, theta, rho):
        '''
        model.angle_to_cortex(theta, rho) yields a (k x 2) matrix in which each row corresponds to
        an area map (e.g. V2 or MT) and the two columns represent x and y coordinates of the
        predicted location in a flattened cortical map at which one would find the visual angle
        values of theta (polar angle) and rho (eccentricity). Theta should be in units of degrees
        between 0 (upper vertical meridian) and 180 (lower vertical meridian).
        If theta and rhos are both vectors of length n, then the result is an (n x k x 2) matrix
        with one entry for each theta/rho pair.
        '''
        raise NotImplementedError(
            'Object with base class RetinotopyModel did not override angle_to_cortex')
    def cortex_to_angle(self, x, y):
        '''
        model.cortex_to_angle(x, y) yields a vector of (polar-angle, eccentricity, id) corresponding
        to the given (x,y) coordinates from a cortical map. The id that is returned is a positive
        integer corresponding to an ROI label (e.g., V1, MT).
        If x and y are vectors of length n then the return value is an (n x 3) matrix in which each
        row corresponds to one (x,y) pair.
        '''
        raise NotImplementedError(
            'Object with base class RetinotopyModel did not override cortex_to_angle')

# How we construct a Schira Model:
@pimms.immutable
class SchiraModel(RetinotopyModel):
    '''
    The SchiraModel class is a class that inherits from RetinotopyModel and acts as a Python wrapper
    around the Java nben.neuroscience.SchiraModel class; it handles conversion from visual field 
    angle to cortical surface coordinates and vice versa for the Banded Double-Sech model proposed
    in the following paper:
    Schira MM, Tyler CW, Spehar B, Breakspear M (2010) Modeling Magnification and Anisotropy in the
    Primate Foveal Confluence. PLOS Comput Biol 6(1):e1000651. doi:10.1371/journal.pcbi.1000651.
    '''

    # These are the accepted arguments to the model:
    default_parameters = pyr.pmap({'A': 1.05,
                                   'B': 90.0,
                                   'lam': 0.4,
                                   'psi': 0.15,
                                   'scale': (21.0, 21.0),
                                   'shear': ((1.0, 0.0), (0.0, 1.0)),
                                   'center': (-6.0, 0.0),
                                   'v1size': 1.2,
                                   'v2size': 0.6,
                                   'v3size': 0.4,
                                   'hv4size': 0.9,
                                   'v3asize': 0.9})
    visual_area_names = ('V1', 'V2', 'V3')

    def __init__(self, A=1.05, B=90.0, lam=0.4, psi=0.15, scale=(21.0, 21.0),
                 shear=((1.0,0.0),(0.0,1.0)), center=(-6.0, 0.0),
                 v1size=1.2, v2size=0.6, v3size=0.4, hv4size=0.9, v3asize=0.9):
        self.area_name_to_id = visual_area_names
        self.parameters = pyr.m(A=A, B=B, lam=lam, psi=psi, scale=scale, shear=shear,
                                center=center, v1size=v1size, v2size=v2size, v3size=v3size,
                                hv4size=hv4size, v3asize=v3asize)
    @pimms.param
    def parameters(params):
        '''
        mdl.parameters is a persistent map of the parameters for the given SchiraModel object mdl.
        '''
        if not pimms.is_pmap(params): params = pyr.pmap(params)
        # do the translations that we need...
        scale = params['scale']
        if pimms.is_number(scale):
            params = params.set('scale', (scale, scale))
        elif not is_tuple(scale):
            params = params.set('scale', tuple(scale))
        shear = params['shear']
        if pimms.is_number(shear) and np.isclose(shear, 0):
            params = params.set('shear', ((1, 0), (0, 1)))
        elif shear[0][0] != 1 or shear[1][1] != 1:
            raise RuntimeError('shear matrix diagonal elements must be 1!')
        elif not is_tuple(shear) or not all(is_tuple(s) for s in shear):
            params.set('shear', tuple([tuple(s) for s in shear]))
        center = params['center']
        if pimms.is_number(center) and np.isclose(center, 0):
            params = params.set('center', (0.0, 0.0))
        return pimms.persist(params, depth=None)

    @pimms.value
    def _java_object(parameters):
        '''
        mdl._java_object is the java representation of the SchiraModel object mdl.
        '''
        # Okay, let's construct the object...
        return java_link().jvm.nben.neuroscience.SchiraModel(
            parameters['A'],
            parameters['B'],
            parameters['lam'],
            parameters['psi'],
            parameters['v1size'],
            parameters['v2size'],
            parameters['v3size'],
            parameters['hv4size'],
            parameters['v3asize'],
            parameters['center'][0],
            parameters['center'][1],
            parameters['scale'][0],
            parameters['scale'][1],
            parameters['shear'][0][1],
            parameters['shear'][1][0])

    def angle_to_cortex(self, theta, rho):
        iterTheta = hasattr(theta, '__iter__')
        iterRho = hasattr(rho, '__iter__')
        jarr = None
        if iterTheta and iterRho:
            if len(theta) != len(rho):
                raise RuntimeError('Arguments theta and rho must be the same length!')
            jarr = self._java_object.angleToCortex(to_java_doubles(theta), to_java_doubles(rho))
        elif iterTheta:
            jarr = self._java_object.angleToCortex(to_java_doubles(theta),
                                                   to_java_doubles([rho for t in theta]))
        elif iterRho:
            jarr = self._java_object.angleToCortex(to_java_doubles([theta for r in rho]),
                                                   to_java_doubles(rho))
        else:
            return self._java_object.angleToCortex(theta, rho)
        return np.asarray([[c for c in r] for r in jarr])
    def cortex_to_angle(self, x, y):
        iterX = hasattr(x, '__iter__')
        iterY = hasattr(y, '__iter__')
        jarr = None
        if iterX and iterY:
            if len(x) != len(y):
                raise RuntimeError('Arguments x and y must be the same length!')
            jarr = self._java_object.cortexToAngle(to_java_doubles(x), to_java_doubles(y))
        elif iterX:
            jarr = self._java_object.cortexToAngle(to_java_doubles(x),
                                                   to_java_doubles([y for i in x]))
        elif iterY:
            jarr = self._java_object.cortexToAngle(to_java_doubles([x for i in y]),
                                                   to_java_doubles(y))
        else:
            return self._java_object.cortexToAngle(x, y)
        dat = np.asarray([[c for c in r] for r in jarr])
        a = dat[:,2]
        a = np.round(np.abs(a))
        a[a > 3] = 0
        dat[:,2] = a
        return dat

@pimms.immutable
class RetinotopyMeshModel(RetinotopyModel):
    '''
    RetinotopyMeshModel is a class that represents a retinotopic map or set of retinotopic maps on
    the flattened 2D cortex.
    RetinotopyMeshModel(tris, coords, polar_angle, eccen, areas) yields a retinotopy mesh model
    object in which the given triangle and coordinate matrices form the mesh and the polar_angle,
    eccen, and areas give the appropriate data for each vertex in coords. Note that the areas
    parameter should be 0 on any boundary vertex and an integer labelling the roi for any other
    vertex.
    '''

    def __init__(self, triangles, coordinates, angles, eccens, area_ids, transform=None,
                 area_name_to_id=None):
        self.faces = triangles
        self.cortical_coordinates = coordinates
        self.polar_angles = angles
        self.eccentricities = eccens
        self.visual_areas = area_ids
        self.transform = transform
        self.area_name_to_id = area_name_to_id
            

    @pimms.param
    def faces(tris):
        'mdl.faces is the triangle matrix for the given retinotopy mesh model mdl.'
        tris = np.asarray(tris, dtype=np.int)
        if tris.shape[0] != 3: tris = tris.T
        if tris.shape[0] != 3: raise ValueError('triangle matrix must have 3 rows or columns')
        return pimms.imm_array(tris)
    @pimms.param
    def cortical_coordinates(coords):
        '''
        mdl.cortical_coordinates is the coordinate matrix for the given retinotopy mesh model mdl's
        representation of the cortical surface.
        '''
        coords = np.asarray(coords)
        if coords.shape[0] != 2: coords = coords.T
        if coords.shape[0] != 2: raise ValueError('coordinate matrix must have 2 rows or columns')
        return pimms.imm_array(coords)
    @pimms.param
    def polar_angles(angs):
        'mdl.polar_angles is the vector of polar angle values for the given retinotopy mesh model.'
        return pimms.imm_array(angs)
    @pimms.param
    def eccentricities(eccs):
        'mdl.eccentrities is the vector of eccentricity values for the given retinotopy mesh model.'
        return pimms.imm_array(eccs)
    @pimms.param
    def visual_areas(labs):
        'mdl.visual_areas is the vector of visual area labels for the given retinotopy mesh model.'
        return pimms.imm_array(labs)
    @pimms.param
    def transform(tx):
        '''
        mdl.transform is the matrix for the affine transform applied to the coordinates on the
        cortical surface representation
        '''
        if tx is None: return None
        tx = to_affine(tx)
        if np.array_equal(tx, np.eye(3)): return None
        tx.setflags(write=False)
        return tx

    @pimms.value
    def inverse_transform(transform):
        '''
        mdl.inverse_transform is the inverse transform (see RetinotopyMeshModel.transform).
        '''
        if transform is None: return None
        return pimms.imm_array(npla.inv(transform))
    @pimms.value
    def visual_coordinates(polar_angles, eccentricities):
        '''
        mdl.cortical_coordinates is the coordinate matrix for the given retinotopy mesh model mdl's
        representation of the cortical surface.
        '''
        z = eccentricities * np.exp(1j * np.pi/180.0 * (90.0 - polar_angles))
        return pimms.imm_array([z.real, z.imag])
    @pimms.value
    def cleaned_visual_areas(visual_areas, faces):
        '''
        mdl.cleaned_visual_areas is the same as mdl.visual_areas except that vertices with visual
        area values of 0 (boundary values) are given the mode of their neighbors.
        '''
        area_ids = np.array(visual_areas)
        boundaryNeis = {}
        for (b,inside) in [(b, set(inside))
                           for t in faces.T
                           for (bound, inside) in [([i for i in t if area_ids[i] == 0],
                                                    [i for i in t if area_ids[i] != 0])]
                           if len(bound) > 0 and len(inside) > 0
                           for b in bound]:
            if b in boundaryNeis: boundaryNeis[b] |= inside
            else:                 boundaryNeis[b] =  inside
        for (b,neis) in six.iteritems(boundaryNeis):
            area_ids[b] = np.argmax(np.bincount(area_ids[list(neis)]))
        return pimms.imm_array(np.asarray(area_ids, dtype=np.int))
    @pimms.value
    def tess(faces, cortical_coordinates, visual_coordinates,
             polar_angles, eccentricities, cleaned_visual_areas):
        'mdl.tess is the tesselation object for mesh model.'
        props = pimms.itable({'polar_angle':  polar_angles,
                              'eccentricity': eccentricities,
                              'visual_area':  cleaned_visual_areas,
                              'cortical_coordinates': cortical_coordinates.T,
                              'visual_coordinates':   visual_coordinates.T})
        if isinstance(faces, geo.Tesselation): return faces.copy(properties=props)
        return geo.Tesselation(faces, properties=props).persist()
    @pimms.value
    def cortical_mesh(tess, cortical_coordinates):
        '''
        mdl.cortical_mesh is the mesh object that represents the 2D cortical surface of the model.
        '''
        return tess.make_mesh(cortical_coordinates).persist()
    @pimms.value
    def visual_meshes(tess, visual_coordinates, cleaned_visual_areas):
        '''
        mdl.visual_meshes is a map of meshes; the keys of the map are the unique visual area id's
        in the given retinotopy mesh model (mdl) and the values are the meshes that represent them.
        '''
        visual_areas = cleaned_visual_areas
        def _make_submesh(area_label):
            def _fn():
                idx = np.where(visual_areas == area_label)[0]
                st = tess.subtess(idx)
                return st.make_mesh(visual_coordinates[:, st.labels]).persist()
            return _fn
        return pimms.lazy_map({k:_make_submesh(k) for k in np.unique(visual_areas) if k != 0})

    
    def cortex_to_angle(self, x, y):
        'See RetinotopyModel.cortex_to_angle.'
        if not pimms.is_vector(x): return self.cortex_to_angle([x], [y])[0]
        # start by applying the transform to the points
        tx = self.inverse_transform
        xy = np.asarray([x,y]).T if tx is None else np.dot(tx, [x,y,np.ones(len(x))])[0:2].T
        # we only need to interpolate from the inverse mesh in this case
        interp_ae = self.cortical_mesh.interpolate(xy, [self.polar_angles, self.eccentricities],
                                                   method='linear')
        interp_id = self.cortical_mesh.interpolate(xy, self.visual_areas,
                                                   method='heaviest')
        interp = np.asarray([interp_ae[0], interp_ae[1], interp_id])
        bad = np.where(np.isnan(np.prod(interp, axis=0)))[0]
        interp[:,bad] = 0.0
        return interp
    def angle_to_cortex(self, theta, rho):
        'See help(neuropythy.registration.RetinotopyModel.angle_to_cortex).'
        #TODO: This should be made to work correctly with visual area boundaries: this could be done
        # by, for each area (e.g., V2) looking at its boundaries (with V1 and V3) and flipping the
        # adjacent triangles so that there is complete coverage of each hemifield, guaranteed.
        if not pimms.is_vector(theta): return self.angle_to_cortex([theta], [rho])[0]
        theta = np.asarray(theta)
        rho = np.asarray(rho)
        zs = np.asarray(
            rho * np.exp([np.complex(z) for z in 1j * ((90.0 - theta)/180.0*np.pi)]),
            dtype=np.complex)
        coords = np.asarray([zs.real, zs.imag]).T
        if coords.shape[0] == 0: return np.zeros((0, len(self.visual_meshes), 2))
        # we step through each area in the forward model and return the appropriate values
        tx = self.transform
        res = np.transpose(
            [self.visual_meshes[area].interpolate(coords, 'cortical_coordinates', method='linear')
             for area in sorted(self.visual_meshes.keys())],
            (1,0,2))
        if tx is not None:
            res = np.asarray(
                [np.dot(tx, np.vstack((area_xy.T, np.ones(len(area_xy)))))[0:2].T
                 for area_xy in res])
        return res

@pimms.immutable
class RegisteredRetinotopyModel(RetinotopyModel):
    '''
    RegisteredRetinotopyModel is a class that represents a retinotopic map or set of retinotopic
    maps on the flattened 2D cortex OR on the 3D cortical surface, via a registration and set of
    map projection parameters.
    RegisteredRetinotopyModel(model, projection_params) yields a retinotopy
    mesh model object in which the given RetinotopyModel object model describes the 2D 
    retinotopy that is predicted for the vertices that result from a map projection, defined using
    the given projection_params dictionary, of the given registration. In other words, the resulting
    RegisteredRetinotopyModel will, when given a hemisphere object to which to apply the model,
    will look up the appropriate registration (by the name registration_name) make a map projection
    of using the given projection_params dictionary, and apply the model to the resulting
    coordinates.
    '''

    def __init__(self, model, mapproj):
        '''
        RegisteredRetinotopyModel(retinotopy_model, map_projection) yields a retinotopy mesh model
        object in which the given RetinotopyModel object describes the 2D retinotopy that is
        predicted for the vertices that result from a map projection, defined using the given
        MapProjection object, of the given registration. In other words, the resulting
        RegisteredRetinotopyModel will, when given a cortex object to which to apply the model, will
        look up the appropriate registration (found in the map projection) make a map projection of
        the cortex using the projection, and apply the model to the resulting coordinates. See also
        neuropythy.geometry.MapProjection.
        '''
        self.model = model
        self.map_projection = mapproj
        self.area_name_to_id = model.area_name_to_id
    @pimms.param
    def model(mdl):
        '''
        rrm.model is the retinotopy model object for the RegisteredRetinotopyModel object rrm.
        '''
        if not isinstance(mdl, RetinotopyModel):
            raise ValueError('given parameter model must be a RetinotopyModel instance')
        return pimms.persist(mdl)
    @pimms.param
    def map_projection(mp):
        '''
        rrm.map_projection is the MapProjection object for the RegisteredRetinotopyModel object rrm.
        '''
        if not isinstance(mp, geo.MapProjection):
            raise ValueError('given parameter map_projection must be a MapProjection instance')
        return pimms.persist(mp)

    def save(self, f):
        '''
        model.save(filename) saves an FMM-formatted file to the given filename; the FMM format is
        used by neuropythy to save and load registered retinotopy models; it can be loaded with the
        load_fmm_model function.
        model.save(file) will write the text directly to the given file.
        '''
        if not isinstance(self.model, RetinotopyMeshModel):
            raise ValueError('Only RetinotopyMeshModels can be saved to an fmm file')
        if pimms.is_str(f):
            with open(f, 'w') as fl:
                self.save(fl)
            return f
        m = self.model
        x0 = self.map_projection.center
        x1 = self.map_projection.center_right
        tx = np.eye(3) if m.transform is None else m.transform
        chir = self.map_projection.chirality
        if chir is not None: chir = chir.upper()
        for ln in ['Flat Mesh Model Version: 1.0',
                   'Points: %d' % m.coordinates.shape[1],
                   'Triangles: %d' % m.faces.shape[1],
                   'Registration: %s' % self.map_projection.registration,
                   'Hemisphere: %s' % chir,
                   'Center: %f,%f,%f' % (x0[0], x0[1], x0[2]),
                   'OnXAxis: %f,%f,%f' % (x1[0], x1[1], x1[2]),
                   'Method: %s' % self.map_projection.method.capitalize(),
                   'Transform: [%f,%f,%f;%f,%f,%f;%f,%f,%f]' % tuple(tuple(x) for x in tx)]:
            f.write(ln + '\n')
        if self.area_name_to_id:
            lbls = [x for (_,x) in sorted(six.iteritems(self.area_name_to_id), key=lambda x:x[0])]
            f.write('AreaNames: [%s]\n' % ' '.join(lbls))
        (xs,ys) = m.coordinates
        for (x,y,t,r,a) in zip(xs, ys, m.polar_angles, m.eccentricities, m.visual_areas):
            f.write('%f,%f :: %f,%f,%f\n' % (x,y,t,r,a))
        for (a,b,c) in zip(**(m.faces + 1)):
            f.write('%d,%d,%d\n' % (a,b,c))
        return f
    def cortex_to_angle(self, *args):
        '''
        The cortex_to_angle method of the RegisteredRetinotopyModel class is identical to that
        of the RetinotopyModel class, but the method may be given a map, mesh, or cortex, in
        which case the result is applied to the coordinates after the appropriate transformation (if
        any) is first applied.
        '''
        if len(args) == 1:
            # see if we can cast it to something; first a mesh...
            try:              m = geo.to_mesh(args[0])
            except Exception: m = None
            if m is not None:
                if geo.is_flatmap(m):
                    (x,y) = m.coordinates
                    return self.model.cortex_to_angle(x, y)
                else:
                    fm = self.map_projection(m)
                    res = np.zeros((3, args[0].vertex_count))
                    c2a = np.asarray(self.cortex_to_angle(fm.coordinates))
                    res[:, fm.labels] = (c2a if len(c2a) == len(res) else c2a.T)
                    return res
            # next, try a cortex
            try:              c = mri.to_cortex(args[0])
            except Exception: c = None
            if c is not None:
                m = self.map_projection(c)
                res = np.zeros((3, c.vertex_count))
                c2a = np.asarray(self.cortex_to_angle(m.coordinates))
                res[:, m.labels] = (c2a if len(c2a) == len(res) else c2a.T)
                return res
            # finally, assume a coordinate matrix
            X = np.asarray(args[0])
            if len(X.shape) != 2: raise ValueError('given coordinate matrix must be rectangular')
            X = X if X.shape[0] == 2 or X.shape[0] == 3 else X.T
            if X.shape[0] == 2: return self.model.cortex_to_angle(X[0], X[1])
            elif X.shape[0] == 3:
                (x,y) = self.map_projection(X)
                return self.model.cortex_to_angle(x, y)
            else: raise ValueError('coordinate matrix must be 2 or 3 dimensional')
        else: return self.model.cortex_to_angle(*args)
    def angle_to_cortex(self, *args):
        '''
        The angle_to_cortex method of the RegisteredRetinotopyModel class is identical to that
        of the RetinotopyModel class, but the method may be given a map, mesh, or hemisphere, in
        which case the result is applied to the 'polar_angle' and 'eccentricity' properties.
        '''
        if len(args) == 1:
            if geo.is_vset(args[0]):
                ang = vis.retinotopy_data(args[0], 'polar_angle')
                ecc = vis.retinotopy_data(args[0], 'eccentricity')
                return self.model.angle_to_cortex(ang, ecc)
            else:
                tr = np.asarray(args)
                if tr.shape[1] == 2: tr = tr.T
                elif tr.shape[0] != 2: raise ValueError('cannot interpret argument')
                return self.model.angle_to_cortex(tr[0], tr[1])
        else: return self.model.angle_to_cortex(*args)

@importer('flatmap_model', ('fmm', 'fmm.gz'))
def load_fmm_model(filename, radius=np.pi/3.0, sphere_radius=100.0):
    '''
    load_fmm_model(filename) yields the fmm model indicated by the given file name. Fmm models are
    triangle meshes that define a field value at every vertex as well as the parameters of a
    projection to the 2D cortical surface via a registartion name and projection parameters.
    The following options may be given:
      * radius (default: pi/3) specifies the radius that should be assumed by the model; fmm models
        do not specify radius values in their projection parameters, but pi/3 is usually large for
        localized models.
      * sphere_radius (default: 100) specifies the radius of the sphere that should be assumed by
        the model. Note that in Freesurfer, spheres have a radius of 100.
    '''
    if not os.path.exists(filename):
        models_path = os.path.join(library_path(), 'models')
        # we look for it in a number of ways:
        fname = next((fnm
                      for beg in ['', models_path] for end in ['', '.fmm', '.fmm.gz']
                      for fnm0 in [filename + end]
                      for fnm in [fnm0 if beg is None else os.path.join(beg, fnm0)]
                      if os.path.exists(fnm)),
                     None)
        if fname is None:
            raise ValueError('Given model/file name does not exist: %s' % filename)
        filename = fname
    if not os.path.isfile(filename):
        raise ValueError('Given filename (%s) is not a file!' % filename)
    gz = True if len(filename) > 3 and filename[-3:] == '.gz' else False
    lines = None
    with (gzip.open(filename, 'rt') if gz else open(filename, 'rt')) as f:
        lines = f.read().split('\n')
    if len(lines) < 3 or lines[0] != 'Flat Mesh Model Version: 1.0':
        raise ValueError('Given file does not contain to a valid flat mesh model!')
    n = int(lines[1].split(':')[1].strip())
    m = int(lines[2].split(':')[1].strip())
    reg = lines[3].split(':')[1].strip()
    hemi = lines[4].split(':')[1].strip().upper()
    center = list(map(float, lines[5].split(':')[1].strip().split(',')))
    onxaxis = list(map(float, lines[6].split(':')[1].strip().split(',')))
    method = lines[7].split(':')[1].strip().lower()
    tx = np.asarray(
        [list(map(float, row.split(',')))
         for row in lines[8].split(':')[1].strip(' \t[]').split(';')])
    if lines[9].startswith('AreaNames: ['):
        # we load the area names
        s = lines[9][12:-1]
        area_names = tuple(s.split(' '))
        l0 = 10        
    else:
        area_names = None
        l0 = 9
    crds = []
    for row in lines[l0:(n+l0)]:
        (left,right) = row.split(' :: ')
        crds.append(list(map(float, left.split(','))))
    crds = np.asarray([list(map(float, left.split(',')))
                       for row in lines[l0:(n+l0)]
                       for (left,right) in [row.split(' :: ')]])
    vals = np.asarray([list(map(float, right.split(',')))
                       for row in lines[l0:(n+l0)]
                       for (left,right) in [row.split(' :: ')]])
    tris = -1 + np.asarray(
        [list(map(int, row.split(',')))
         for row in lines[(n+l0):(n+m+l0)]])
    return RegisteredRetinotopyModel(
        RetinotopyMeshModel(tris, crds,
                            90-180/np.pi*vals[:,0], vals[:,1], np.asarray(vals[:,2], dtype=np.int),
                            transform=tx,
                            area_name_to_id=area_names),
        geo.MapProjection(registration=reg,
                          center=center,
                          center_right=onxaxis,
                          method=method,
                          radius=radius,
                          sphere_radius=sphere_radius,
                          chirality=hemi))



####################################################################################################
# neuropythy/hcp/core.py
# Simple tools for use with the Human Connectome Project (HCP) database in Python
# By Noah C. Benson

import numpy                        as np
import nibabel                      as nib
import nibabel.freesurfer.io        as fsio
import nibabel.freesurfer.mghformat as fsmgh
import pyrsistent                   as pyr
from   six.moves                import collections_abc as collections
import os, warnings, six, pimms

from ..        import geometry      as geo
from ..        import mri           as mri
from ..        import io            as nyio
from ..        import vision        as nyvis

from ..util    import (library_path, curry, config, to_pseudo_path, is_pseudo_path)
from .files    import (subject_paths, clear_subject_paths, add_subject_path, find_subject_path,
                       subject_file_map, is_hcp_subject_path)

def to_default_alignment_value(x):
    if not pimms.is_str(x): raise ValueError('hcp_default_alignment must be a string')
    x = x.lower()
    if   x in ('msmsulc', 'sulc'):  x = 'MSMSulc'
    elif x in ('msmall', 'all'):    x = 'MSMAll'
    elif x in ('fs', 'freesurfer'): x = 'FS'
    else: raise ValueError('invalid value for hcp_default_alignment: %s')
    return x
config.declare('hcp_default_alignment', environ_name='HCP_DEFAULT_ALIGNMENT',
               filter=to_default_alignment_value, default_value='MSMAll')
def cortex_from_filemap(fmap, name, affine=None):
    '''
    cortex_from_filemap(filemap, name) yields a cortex object from the given filemap.
    '''
    chirality = name[:2].lower()
    # get the relevant hemi-data
    hdat = fmap.data_tree.hemi[name]
    # we need the tesselation at build-time, so let's create that now:
    tris = hdat.tess['white']
    # this tells us the max number of vertices
    n = np.max(tris) + 1
    # Properties: we want to merge a bunch of things together...
    # for labels, weights, annots, we need to worry about loading alts:
    def _load_with_alt(k, s0, sa, trfn):
        if s0 is not None:
            try: u = s0.get(k, None)
            except Exception: u = None
        else: u = None
        if u is None and sa is not None:
            try: u = sa.get(k, None)
            except Exception: u = None
        if u is None: raise ValueError('Exception while loading property %s' % k)
        else: return u if trfn is None else trfn(u)
    def _lbltr(ll):
        l = np.zeros(n, dtype='bool')
        l[ll[0]] = True
        l.setflags(write=False)
        return l
    def _wgttr(ll):
        w = np.zeros(n, dtype='float')
        w[ll[0]] = ll[1]
        w.setflags(write=False)
        return w
    def _anotr(ll):
        ll[0].setflags(write=False)
        return ll[0]
    p = {}
    from itertools import chain
    l = hdat.label if hasattr(hdat, 'label') else {}
    al = hdat.alt_label if hasattr(hdat, 'alt_label') else {}
    for k in set(chain(six.iterkeys(l), six.iterkeys(al))):
        p[k+'_label'] = curry(_load_with_alt, k, l, al, _lbltr)
    w = hdat.weight if hasattr(hdat, 'weight') else {}
    aw = hdat.alt_weight if hasattr(hdat, 'alt_weight') else {}
    for k in set(chain(six.iterkeys(w), six.iterkeys(aw))):
        p[k+'_weight'] = curry(_load_with_alt, k, w, aw, _wgttr)
    a = hdat.annot if hasattr(hdat, 'annot') else {}
    aa = hdat.alt_annot if hasattr(hdat, 'alt_annot') else {}
    for k in set(chain(six.iterkeys(a), six.iterkeys(aa))):
        p[k] = curry(_load_with_alt, k, a, aa, _anotr)
    props = pimms.merge(hdat.property if hasattr(hdat, 'property') else {}, pimms.lazy_map(p))
    tess = geo.Tesselation(tris, properties=props)
    # if this is a subject that exists in the library, we may want to add some files:
    if name is None:
        pd = fmap.pseudo_paths[None]._path_data
        name = pd['pathmod'].split(fmap.actual_path)[1]
    regs = hdat.registration if hasattr(hdat, 'registration') else {}
    # Okay, make the cortex object!
    md = {'file_map': fmap}
    if name is not None: md['subject_id'] = name
    srfs = hdat.surface if hasattr(hdat, 'surface') else {}
    return mri.Cortex(chirality, tess, srfs, regs, affine=affine, meta_data=md).persist()
def images_from_filemap(fmap):
    '''
    images_from_filemap(fmap) yields a persistent map of MRImages tracked by the given subject with
      the given name and path; in freesurfer subjects these are renamed and converted from their
      typical freesurfer filenames (such as 'ribbon') to forms that conform to the neuropythy naming
      conventions (such as 'gray_mask'). To access data by their original names, use the filemap.
    '''
    imgmap = fmap.data_tree.image
    def img_loader(k): return lambda:imgmap[k]
    imgs = {k:img_loader(k) for k in six.iterkeys(imgmap)}
    def _make_mask(val, eq=True):
        rib = imgmap['ribbon']
        img = np.asarray(rib.dataobj)
        arr = (img == val) if eq else (img != val)
        arr.setflags(write=False)
        return type(rib)(arr, rib.affine, rib.header)
    imgs['lh_gray_mask']  = lambda:_make_mask(3)
    imgs['lh_white_mask'] = lambda:_make_mask(2)
    imgs['rh_gray_mask']  = lambda:_make_mask(42)
    imgs['rh_white_mask'] = lambda:_make_mask(41)
    imgs['brain_mask']    = lambda:_make_mask(0, False)
    # merge in with the typical images
    return pimms.merge(fmap.data_tree.image, pimms.lazy_map(imgs))
def subject_from_filemap(fmap, name=None, meta_data=None, check_path=True,
                         default_alignment='MSMAll'):
    '''
    subject_from_filemap(fmap) yields an HCP subject from the given filemap.
    '''
    # start by making a pseudo-dir:
    if check_path and not is_hcp_subject_path(fmap.pseudo_paths[None]):
        raise ValueError('given path does not appear to hold an HCP subject')
    # we need to go ahead and load the ribbon...
    rib = fmap.data_tree.image['ribbon']
    vox2nat = rib.affine
    # make images and hems
    imgs = images_from_filemap(fmap)
    # many hemispheres to create:
    hems = pimms.lazy_map({h:curry(cortex_from_filemap, fmap, h)
                           for h in ['lh_native_MSMAll',  'rh_native_MSMAll',
                                     'lh_nat32k_MSMAll',  'rh_nat32k_MSMAll',
                                     'lh_nat59k_MSMAll',  'rh_nat59k_MSMAll',
                                     'lh_LR32k_MSMAll',   'rh_LR32k_MSMAll',
                                     'lh_LR59k_MSMAll',   'rh_LR59k_MSMAll',
                                     'lh_LR164k_MSMAll',  'rh_LR164k_MSMAll',
                                     'lh_native_MSMSulc', 'rh_native_MSMSulc',
                                     'lh_nat32k_MSMSulc', 'rh_nat32k_MSMSulc',
                                     'lh_nat59k_MSMSulc', 'rh_nat59k_MSMSulc',
                                     'lh_LR32k_MSMSulc',  'rh_LR32k_MSMSulc',
                                     'lh_LR59k_MSMSulc',  'rh_LR59k_MSMSulc',
                                     'lh_LR164k_MSMSulc', 'rh_LR164k_MSMSulc',
                                     'lh_native_FS',      'rh_native_FS',
                                     'lh_nat32k_FS',      'rh_nat32k_FS',
                                     'lh_nat59k_FS',      'rh_nat59k_FS',
                                     'lh_LR32k_FS',       'rh_LR32k_FS',
                                     'lh_LR59k_FS',       'rh_LR59k_FS',
                                     'lh_LR164k_FS',      'rh_LR164k_FS']})
    # now, setup the default alignment aliases:
    if default_alignment is not None:
        for h in ['lh_native',  'rh_native',  'lh_nat32k',  'rh_nat32k',
                  'lh_nat59k',  'rh_nat59k',  'lh_LR32k',   'rh_LR32k',
                  'lh_LR59k',   'rh_LR59k',   'lh_LR164k',  'rh_LR164k']:
            hems = hems.set(h, curry(lambda h:hems[h+'_'+default_alignment], h))
        hems = hems.set('lh', lambda: hems['lh_native'])
        hems = hems.set('rh', lambda: hems['rh_native'])
    meta_data = pimms.persist({} if meta_data is None else meta_data)
    meta_data = meta_data.set('raw_images', fmap.data_tree.raw_image)
    if default_alignment is not None:
        meta_data = meta_data.set('default_alignment', default_alignment)
    return mri.Subject(name=name, pseudo_path=fmap.pseudo_paths[None],
                       hemis=hems, images=imgs,
                       meta_data=meta_data).persist()
@nyio.importer('hcp_subject', sniff=is_hcp_subject_path)
def subject(path, name=Ellipsis, meta_data=None, check_path=True, filter=None,
            default_alignment=Ellipsis):
    '''
    subject(name) yields an HCP-based Subject object for the subject with the given name or path.
      Subjects are cached and not reloaded, so multiple calls to subject(name) will yield the same
      immutable subject object.

    The name argument is allowed to take a variety of forms:
      * a local (absolute or relative) path to a valid HCP subject directory
      * a url or pseudo-path to a valid HCP subject
      * an integer, in which case the neuropythy.data['hcp'] dataset is used (i.e., the subject data
        are auto-downloaded from the HCP Amazon-S3 bucket as required)
    If you request a subject by path, the HCP module has no way of knowing for sure if that subject
    should be auto-downloaded, so is not; if you want subjects to be auto-downloaded from the HCP
    database, you should represent the subjects by their integer ids.

    Note that subects returned by hcp_subject() are always persistent Immutable objects; this
    means that you must create a transient version of the subject to modify it via the member
    function sub.transient(). Better, you can make copies of the objects with desired modifications
    using the copy method--see the pimms library documentation regarding immutable classes and
    objects.

    If you wish to modify all subjects loaded by this function, you may set its attribute 'filter'
    to a function or list of functions that take a single argument (a subject object) and returns a
    single argument (a potentially-modified subject object).

    The argument name may alternately be a pseudo_path object or a path that can be converted into a
    pseudo_path object.

    The following options are accepted:
      * name (default: Ellipsis) may optionally specify the subject's name; if Ellipsis, then
        attempts to deduce the name from the initial argument (which may be a name or a path).
      * meta_data (default: None) may optionally be a map that contains meta-data to be passed along
        to the subject object (note that this meta-data will not be cached).
      * check_path (default: True) may optionally be set to False to ignore the requirement that a
        directory contain at least the mri/, label/, and surf/ directories to be considered a valid
        HCP subject directory. Subject objects returned when this argument is not True are not
        cached. Additionally, check_path may be set to None instead of False, indicating that no
        sanity checks or search should be performed whatsoever: the string name should be trusted 
        to be an exact relative or absolute path to a valid HCP subejct.
      * filter (default: None) may optionally specify a filter that should be applied to the subject
        before returning. This must be a function that accepts as an argument the subject object and
        returns a (potentially) modified subject object. Filtered subjects are cached by using the
        id of the filters as part of the cache key.
      * default_alignment (default: Ellipsis) specifies the default alignment to use with HCP
        subjects; this may be either 'MSMAll', 'MSMSulc', or 'FS'; the deafult (Ellipsis) indicates
        that the 'hcp_default_alignment' configuration value should be used (by default this is
        'MSMAll').
    '''
    from neuropythy import data
    if pimms.is_str(default_alignment):
        default_alignment = to_default_alignment_value(default_alignment)
    elif default_alignment in (Ellipsis, None):
        default_alignment = config['hcp_default_alignment']
    # first thing: if the sid is an integer, we try to get the subject from the hcp dataset;
    # in this case, because the hcp dataset actually calls down through here (with a pseudo-path),
    # we don't need to run any of the filters that are run below (they have already been run)
    if pimms.is_int(path):
        if default_alignment != 'MSMAll':
            warnings.warn('%s alignment requested, but MSMAll used for HCP release subject'
                          % (default_alignment,))
        try: return data['hcp'].subjects[path]
        except Exception: pass
    # convert the path to a pseudo-dir; this may fail if the user is requesting a subject by name...
    try: pdir = to_pseudo_path(path)
    except Exception: pdir = None
    if pdir is None: # search for a subject with this name
        tmp = find_subject_path(path, check_path=check_path)
        if tmp is not None:
            pdir = to_pseudo_path(tmp)
            path = tmp
    if pdir is None:
        if default_alignment != 'MSMAll':
            warnings.warn('%s alignment requested, but MSMAll used for HCP release subject'
                          % (default_alignment,))
        # It's possible that we need to check the hcp dataset
        try: return data['hcp'].subjects[int(path)]
        except: pass
        raise ValueError('could not find HCP subject: %s' % (path,))
    path = pdir.actual_source_path
    # okay, before we continue, lets check the cache...
    tup = (path, default_alignment)
    if tup in subject._cache: sub = subject._cache[tup]
    else:
        # extract the name if need-be
        if name is Ellipsis:
            import re
            (pth,name) = (path, '.')
            while name == '.': (pth, name) = pdir._path_data['pathmod'].split(pth)
            name = name.split(':')[-1]
            name = pdir._path_data['pathmod'].split(name)[1]
            if '.tar' in name: name = name.split('.tar')[0]
        # make the filemap
        fmap = subject_file_map(pdir, name=name)
        # and make the subject!
        sub = subject_from_filemap(fmap, name=name, check_path=check_path, meta_data=meta_data,
                                   default_alignment=default_alignment)
        if mri.is_subject(sub):
            sub = sub.persist()
            sub = sub.with_meta(file_map=fmap)
            subject._cache[(path, default_alignment)] = sub
    # okay, we have the initial subject; let's organize the filters
    if pimms.is_list(subject.filter) or pimms.is_tuple(subject.filter): filts = list(subject.filter)
    else: filts = []
    if pimms.is_list(filter) or pimms.is_tuple(filter): filter = list(filter)
    else: filter = []
    filts = filts + filter
    if len(filts) == 0: return sub
    fids = tuple([id(f) for f in filts])
    tup = fids + (path, default_alignment)
    if tup in subject._cache: return subject._cache[tup]
    for f in filts: sub = f(sub)
    if mri.is_subject(sub): subject._cache[tup] = sub
    return sub.persist()
subject._cache = {}
subject.filter = None

def forget_subject(sid):
    '''
    forget_subject(sid) causes neuropythy's hcp module to forget about cached data for the subject
      with subject id sid. The sid may be any sid that can be passed to the subject() function.
    
    This function is useful for batch-processing of subjects in a memory-limited environment; e.g.,
    if you run out of memory while processing hcp subjects it is possibly because neuropythy is
    caching all of their data instead of freeing it.
    '''
    sub = subject(sid)
    if sub.path in subject._cache:
        del subject._cache[sub.path]
    else:
        for (k,v) in six.iteritems(subject._cache):
            if v is sub:
                del subject._cache[k]
                break
    return None
def forget_all():
    '''
    forget_all() causes neuropythy's hcp module to forget all cached subjects. See also
    forget_subject.
    '''
    subject._cache = {}
    return None
    
def download(sid):
    '''
    neuropythy.hcp.download(sid) is equivalent to neuropythy.data['hcp'].download(sid).
    '''
    import neuropythy as ny
    return ny.data['hcp'].download(sid)

# This is copied from ny.freesurfer.core; changes here should be duplicated there. This isn't really
# a good way to organize this--#TODO is to unify the subject-path interfaces somehow.
class SubjectDir(collections.Mapping):
    '''
    SubjectsDir objects are dictionary-like objects that represent a particular subject directory.
    They satisfy their queries (such as `111312 in spath`) by querying the filesystem itself.

    For more information see the subjects_path function.
    '''
    def __init__(self, path, bids=False, filter=None, meta_data=None, check_path=True):
        path = os.path.expanduser(os.path.expandvars(path))
        self.path = os.path.abspath(path)
        if not os.path.isdir(self.path): raise ValueError('given path is not a directory')
        self.bids = bool(bids)
        self.options = dict(filter=filter, meta_data=meta_data, check_path=bool(check_path))
    def __contains__(self, sub):
        if pimms.is_int(sub): sub = str(sub)
        # first check the item straight-up:
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_hcp_subject_path(sd)): return True
        if not bids: return False
        if sub.startswith('sub-'): sd = os.path.join(self.path, sub[4:])
        else: sd = os.path.join(self.path, 'sub-' + sub)
        return os.path.isdir(sd) and (not check_path or is_hcp_subject_path(sd))
    def _get_subject(self, sd, name):
        try:
            if name == str(int(name)): name = int(name)
        except Exception: pass
        opts = dict(**self.options)
        opts['name'] = name
        return subject(sd, **opts)
    def __getitem__(self, sub):
        if pimms.is_int(sub): sub = str(sub)
        check_path = self.options['check_path']
        if self.bids:
            if sub.startswith('sub-'): (sub, name) = (sub, name[4:])
            else: (sub,name) = ('sub-' + sub, sub)
        else: name = sub
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_hcp_subject_path(sd)):
            return self._get_subject(sd, name)
        if not bids: return False
        # try without the 'sub-' (e.g. for fsaverage)
        sub = name
        sd = os.path.join(self.path, sub)
        if os.path.isdir(sd) and (not check_path or is_hcp_subject_path(sd)):
            return self._get_subject(sd, name)
        raise KeyError(sub)
    def asmap(self):
        check_path = self.options['check_path']
        # search through the subjects in this directory
        res = {}
        for sub in os.listdir(self.path):
            if self.bids and sub.startswith('sub-'): (sub,name) = (sub,sub[4:])
            else: name = sub
            sdir = os.path.join(self.path, sub)
            if not check_path or is_hcp_subject_path(sdir):
                res[name] = curry(self._get_subject, sdir, name)
        return pimms.lmap(res)
    def __len__(self): return len(self.asmap())
    def __iter__(self): return iter(self.asmap())
    def __repr__(self): return 'freesurfer.SubjectsPath(' + repr(self.asmap()) + ')'
    def __str__(self): return 'freesurfer.SubjectsPath(' + str(self.asmap()) + ')'
# Functions for handling freesurfer subject directories themselves
def is_hcp_subject_dir_path(path):
    '''
    is_hcp_subject_dir_path(path) yields True if path is a directory that contains at least
      one HCP subejct subdirectory.
    '''
    if not os.path.isdir(path): return False
    for p in os.listdir(path):
        pp = os.path.join(path, p)
        if not os.path.isdir(pp): continue
        if is_hcp_subject_path(pp): return True
    return False
@nyio.importer('hcp_subject_dir', sniff=is_hcp_subject_dir_path)
def subject_dir(path, bids=False, filter=None, meta_data=None, check_path=True):
    '''
    subject_dir(path) yields a dictionary-like object containing the subjects in the FreeSurfer
      subjects directory given by path.

    The following optional arguments are accepted:
      * bids (default: False) may be set to True to indicate that the directory is part of a BIDS
        directory; in this case the 'sub-' prefix is stripped from directory names to deduce subject
        names.
      * check_path (default: True) may optionally be set to False to ignore the requirement that a
        subject directory contain at least the mri/, label/, and surf/ directories to be considered
        a valid FreeSurfer subject directory. See help(neuropythy.freesurfer.subject) for more
        information.
      * filter (default: None) may optionally specify a filter that should be applied to the subject
        before returning. See help(neuropythy.freesurfer.subject) for more information.
      * meta_data (default: None) may optionally be a map that contains meta-data to be passed along
        to the subject object (note that this meta-data will not be cached).
    '''
    return SubjectDir(path, bids=bids, filter=filter, meta_data=meta_data, check_path=check_path)
    
    



####################################################################################################
# neuropythy/mri/__init__.py
# Data structures and simple tools for dealing with the cortex and cortical data.
# By Noah C. Benson

'''
The neuropythy.mri package contains definitions of the relevant data-structures and various tools
for interacting with cortical data. The primary data include:

  * Cortex, a class based on neuropythy.geometry.Topology, which tracks the various layers of the 
    cortex in the form of CorticalMesh objects.
  * Subject, a class that tracks data connected to an individual subject.
'''

from .core   import (Subject, Cortex, is_subject, is_cortex, to_cortex)
from .images import (to_image_spec, to_image, to_image_header, image_interpolate, image_apply,
                     image_reslice, is_image_spec, image_array_to_spec, image_header_to_spec,
                     image_to_spec, image_copy, image_clear, is_pimage, is_npimage,
                     to_image_type)
from ..util  import (is_image, is_image_header)



####################################################################################################
# neuropythy/hcp/files.py
# Stored data regarding the organization of the files in HCP subjects.
# by Noah C. Benson

import os, six, logging, copy, pimms, pyrsistent as pyr, nibabel as nib, numpy as np
from .. import io as nyio
from ..util import (config, is_image, to_credentials, file_map, to_pseudo_path, is_pseudo_path,
                    curry, library_path)
from ..freesurfer import (freesurfer_subject_filemap_instructions,
                          freesurfer_subject_data_hierarchy)

# this isn't required, but if we can load it we will use it for auto-downloading subject data
try:              import s3fs
except Exception: s3fs = None

####################################################################################################
# Subject Directory and where to find Subjects

  
def to_subject_paths(paths):
    '''
    to_subject_paths(paths) accepts either a string that is a :-separated list of directories or a
      list of directories and yields a list of all the existing directories.
    '''
    if paths is None: return []
    if pimms.is_str(paths): paths = paths.split(':')
    pp = [os.path.expanduser(os.path.expandvars(p)) for p in paths]
    return [p for p in pp if os.path.isdir(p)]
config.declare('hcp_subject_paths', environ_name='HCP_SUBJECTS_DIR', filter=to_subject_paths)
def subject_paths():
    '''
    subject_paths() yields a list of paths to HCP subject directories in which subjects are
      automatically searched for when identified by subject-name only. These paths are searched in
      the order returned from this function.

    If you must edit these paths, it is recommended to use add_subject_path, and clear_subject_paths
    functions.
    '''
    return config['hcp_subject_paths']
def clear_subject_paths(subpaths):
    '''
    clear_subject_paths() resets the HCP subject paths to be empty and yields the previous
      list of subject paths.
    '''
    sd = config['hcp_subject_paths']
    config['hcp_subject_paths'] = []
    return sd
def add_subject_path(path, index=None):
    '''
    add_subject_path(path) will add the given path to the list of subject directories in which to
      search for HCP subjects. The optional argument index may be given to specify the precedence of
      this path when searching for a new subject; the default, 0, always inserts the path at the
      front of the list; a value of k indicates that the new list should have the new path at index
      k.
    The path may contain :'s, in which case the individual directories are separated and added.  If
    the given path is not a directory or the path could not be inserted, yields False; otherwise,
    yields True. If the string contains a : and multiple paths, then True is yielded only if all
    paths were successfully inserted.  See also subject_paths.
    '''
    paths = [p for p in path.split(':') if len(p) > 0]
    if len(paths) > 1:
        tests = [add_subject_path(p, index=index) for p in reversed(paths)]
        return all(t for t in tests)
    else:
        spaths = config['hcp_subject_paths']
        path = os.path.expanduser(path)
        if not os.path.isdir(path): return False
        if path in spaths: return True
        try:
            if index is None or index is Ellipsis:
                sd = spaths + [path]
            else:
                sd = spaths + []
                sd.insert(index, path)
            config['hcp_subject_paths'] = sd
            return True
        except Exception:
            return False
def is_hcp_subject_path(path):
    '''
    is_hcp_subject_path(path) yields True if the given path appears to be an HCP subject directory;
      specifically, the path must be a directory that contains the subdirectories 'MNINonLinear' and
      'T1w'.
    '''
    needed = ['T1w', 'MNINonLinear']
    if   is_pseudo_path(path): return all(path.find(k) is not None for k in needed)
    elif os.path.isdir(path):  return all(os.path.isdir(os.path.join(path, k)) for k in needed)
    else:                      return False
def find_subject_path(sid, check_path=True):
    '''
    find_subject_path(sub) yields the full path of a HCP subject with the name given by the string
      sub, if such a subject can be found in the HCP search paths. See also add_subject_path.

    If no subject is found, then None is returned.
    '''
    # if it's a full/relative path already, use it:
    sub = str(sid)
    if ((not check_path or is_hcp_subject_path(sub)) and
        (check_path is None or os.path.isdir(sub))):
        return sub
    # check the subject directories:
    sdirs = config['hcp_subject_paths']
    return next((os.path.abspath(p) for sd in sdirs
                 for p in [os.path.join(sd, sub)]
                 if ((not check_path or is_hcp_subject_path(p)) and
                     (check_path is None or os.path.isdir(p)))),
                None)

if config['hcp_subject_paths'] is None:
    # if a path wasn't found, there are a couple environment variables we want to look at...
    if 'HCPSUBJS_DIR' in os.environ: add_subject_path(os.environ['HCPSUBJS_DIR'])
    for varname in ['HCP_ROOT', 'HCP_DIR']:
        if varname in os.environ:
            dirname = os.path.join(os.environ[varname], 'subjects')
            if os.path.isdir(dirname):
                add_subject_path(dirname)

####################################################################################################
# Subject Data Structure
# This structure details how neuropythy understands an HCP subject to be structured.

def gifti_to_array(gii):
    '''
    gifti_to_array(gii) yields the squeezed array of data contained in the given gifti object, gii,
      Note that if gii does not contain simple data in its darray object, then this will produce
      undefined results. This operation is effectively equivalent to:
      np.squeeze([x.data for x in gii.darrays]).
    gifti_to_array(gii_filename) is equivalent to gifti_to_array(neyropythy.load(gii_filename)).
    '''
    if pimms.is_str(gii): return gifti_to_array(ny.load(gii, 'gifti'))
    elif pimms.is_nparray(gii): return gii #already done
    elif isinstance(gii, nib.gifti.gifti.GiftiImage):
        return np.squeeze(np.asarray([x.data for x in gii.darrays]))
    else: raise ValueError('Could not understand argument to gifti_to_array')

def cifti_label_data(obj):
    '''
    cifti_label_data(obj) yields an ordered dictionary of label names mapped to label data in the
      given cifti object obj; if obj is not an interpretable label structure, None is returned.
    
    If the argument obj is a CIFTI image object then a list of values, one per axis in the CIFTI
    object, is returned where None indicates that the axis was not a label axis. If obj is an axis,
    then either None or the datatable alone is returned.

    Note that the label data for a particular label name may be empty/None; this indicates that the
    axis is for scalar data rather than integer label data.
    '''
    import nibabel.cifti2 as cii
    from neuropythy import label_index
    if obj is None: return None
    elif isinstance(obj, cii.Cifti2Image):
        return cifti_label_data(obj.header)
    elif isinstance(obj, cii.Cifti2Header):
        return [cifti_label_data(obj.get_index_map(k)) for k in obj.mapped_indices]
    elif not isinstance(obj, cii.Cifti2MatrixIndicesMap):
        return None
    # okay, it's a matrix-indices-map object... see if it has a named map
    nmaps = list(obj.named_maps)
    if len(nmaps) == 0: return None
    # it has a named map and label-table; we just have to interpret it
    from collections import OrderedDict
    res = OrderedDict()
    for nmap in nmaps:
        lbls = {}
        for k in ([] if nmap.label_table is None else nmap.label_table):
            lbl = nmap.label_table[k]
            lbls[lbl.key] = (lbl.label, lbl.rgba)
        if nmap.metadata is None or len(nmap.metadata) == 0: md = pyr.m()
        else: md = pimms.persist(nmap.metadata.data)
        if len(lbls) > 0: lbls = label_index(lbls, meta_data=md)
        else: lbls = None
        res[nmap.map_name] = lbls
    return res
def cifti_axis_spec(ax):
    '''
    cifti_axis_spec(obj) yields a persistent mapping of data about the given cifti2 axis object obj.

    The argument obj may alternately be a cifti image object or an image header, in which case a
    list of axis specs are given, one per axis.

    The map that is returned by this function contains the keys 'axis_type' ('CIFTI_INDEX_TYPE_' +
    one of 'SURFACE', 'VOLUME', 'PARCELS', 'BRAIN_MODELS', 'SERIES', 'SCALARS', or 'LABELS') and
    various data relevant to each specific type.
    '''
    import nibabel.cifti2 as cii
    from neuropythy import label_index
    if ax is None: return None
    elif isinstance(ax, cii.Cifti2Image):
        return cifti_axis_spec(ax.header)
    elif isinstance(ax, cii.Cifti2Header):
        return [cifti_axis_spec(ax.get_index_map(k)) for k in ax.mapped_indices]
    elif not isinstance(ax, cii.Cifti2MatrixIndicesMap):
        return None
    d = {}
    # start by checking for surface and volume data; they are used in parcels and brain models
    sdat = {}
    for srf in ax.surfaces:
        sdat[srf.brain_structure] = {'vertex_count': srf.surface_number_of_vertices}
    if len(sdat) > 0: d['surfaces'] = sdat
    if ax.volume is not None:
        d['volume_shape'] = ax.volume.volume_dimensions
        mtx = ax.volume.transformation_matrix_voxel_indices_ijk_to_xyz.matrix
        # 3 + exp because -3 is mm and we prefer these data to be in mm
        exp = 3 + ax.volume.transformation_matrix_voxel_indices_ijk_to_xyz.meter_exponent
        if exp != 0: mtx = mtx * 10**exp
        d['volume_affine'] = mtx
    # Okay, note the axis type:
    axtype = ax.indices_map_to_data_type
    d['axis_type'] = axtype
    # Check for series data:
    (i0,n,step) = (ax.series_start, ax.number_of_series_points, ax.series_step)
    if n is not None:
        idcs = np.arange(i0, i0 + n*step, step)
        #imsh.append(len(idcs))
        if ax.series_unit is not None:
            try: idcs = pimms.quant(idcs, ax.series_unit.lower())
            except Exception: pass
        d['series'] = idcs
        d['size'] = len(idcs)
    # Check for brain-model data:
    bmdls = list(ax.brain_models)
    if len(bmdls) > 0:
        maxii = 0
        idxdat = {}
        for mdl in sorted(bmdls, key=lambda m:m.index_offset):
            sl = slice(mdl.index_offset, mdl.index_offset + mdl.index_count)
            if maxii < sl.stop: maxii = sl.stop
            if mdl.model_type == 'CIFTI_MODEL_TYPE_SURFACE':
                idxdat[mdl.brain_structure] = {'vertex_count': mdl.surface_number_of_vertices,
                                               'vertex_indices': np.array(mdl.vertex_indices),
                                               'cifti_indices': sl}
            elif mdl.model_type == 'CIFTI_MODEL_TYPE_VOXELS':
                idxdat[mdl.brain_structure] = {'voxel_indices': np.array(mdl.voxel_indices_ijk),
                                               'cifti_indices': sl}
            else: raise ValueError('Unrecognized CIFTI model type: %s' % mdl.model_type)
        d['structures'] = idxdat
        d['size'] = maxii
    # Check for label and scalar data:
    lbls = cifti_label_data(ax)
    if lbls is not None:
        d['size'] = len(lbls)
        d['names'] = list(lbls.keys())
        if not all(v is None for v in six.itervalues(lbls)): d['labels'] = list(lbls.values())
    # Check for parcel data:
    parcs = {}
    maxii = 0
    for p in ax.parcels:
        pp = {}
        if p.voxel_indices_ijk is not None:
            pp['voxels'] = np.array(p.voxel_indices_ijk)
            maxii += len(pp['voxels'])
        vv = {}
        for vrt in p.vertices:
            vv[vrt.brain_structure] = np.array(vrt.vertices)
            maxii += len(vv[vrt.brain_structure])
        if len(vv) > 0: pp['vertices'] = vv
        parcs[p.name] = pp
    if len(parcs) > 0:
        d['parcels'] = parcs
        d['size'] = maxii
    # That's all we check for currently
    return d

def cifti_split(cii, label=('lh', 'rh', 'rest'), subject=None, hemi=None, null=np.nan):
    '''
    cifti_split(cii, label) yields the rows or columns of the given cifti file that correspond to
      the given label (see below).
    cifti_split(cii) is equivalent to cifti_split(cii, ('lh', 'rh', 'rest')).

    The label argument may be any of the following:
      * a valid CIFTI label name such as 'CIFTI_STRUCTURE_CEREBELLUM' or
        'CIFTI_STRUCTURE_CORTEX_LEFT';
      * an abbreviated name such as 'cerebellum' for 'CIFTI_STRUCTURE_CEREBELLUM'.
      * the abbreviations 'lh' and 'rh' which stand for 'CIFTI_STRUCTURE_CORTEX_LEFT' and 
        'CIFTI_STRUCTURE_CORTEX_RIGHT';
      * the special keyword 'rest', which represents all the rows/columns not collected by any other
        instruction ('rest', by itself, results in the whole matrix being returned); or
      * A tuple of the above, indicating that each of the items listed should be returned
        sequentially in a tuple.

    The following optional arguments may be given:
      * subject (default: None) may specify the subject
      * hemi (default: None) can specify the hemisphere object that 
    '''
    if cii is None: raise ValueError('cifti_split() called with argument None')
    dat = np.asanyarray(cii.dataobj if is_image(cii) else cii)
    n = dat.shape[-1]
    atlas = cifti_split._size_data.get(n, None)
    if atlas is None: raise ValueError('cannot split cifti with size %d' % n)
    if atlas not in cifti_split._atlas_cache:
        patt = os.path.join('data', 'fs_LR', '%s.atlasroi.%dk_fs_LR.shape.gii')
        lgii = nib.load(os.path.join(library_path(), patt % ('lh', atlas)))
        rgii = nib.load(os.path.join(library_path(), patt % ('rh', atlas)))
        cifti_split._atlas_cache[atlas] = tuple([pimms.imm_array(gii.darrays[0].data.astype('bool'))
                                                 for gii in (lgii, rgii)])
    (lroi,rroi) = cifti_split._atlas_cache[atlas]
    (ln,lN) = (np.sum(lroi), len(lroi))
    (rn,rN) = (np.sum(rroi), len(rroi))
    (ldat,rdat,sdat) = [np.full(dat.shape[:-1] + (k,), null) for k in [lN, rN, n - ln - rn]]
    ldat[..., lroi] = dat[..., :ln]
    rdat[..., rroi] = dat[..., ln:(ln+rn)]
    sdat[...] = dat[..., (ln+rn):]
    if ln + rn >= n: sdat = None
    return (ldat, rdat, sdat)
cifti_split._size_data = {
    # two sizes for each atlas: one for when the cifti file includes subcortical voxels and one for
    # when it includes only the surface vertices
    91282:  32,
    59412:  32,
    170494: 59,
    108441: 59,
    # not sure what the bigger size is for this...
    #?????: 164,
    298261: 164}
cifti_split._atlas_cache = {}
def cifti_extract(cii, h, null=np.nan):
    '''
    cifti_extract(cii, h) yields the portion of the cifti vector or matrix that is associated
      with the given hemisphere h of the given subject. If h is None, then yields any subcortical
      voxels (or None if none are included).
    '''
    (l,r,s) = cifti_split(cii, null=null)
    if h is None:            return s
    elif h.startswith('lh'): return l
    else:                    return r

####################################################################################################
# The filemap instructions and related helper data/functions

def subject_file_map(path, name=None):
    '''
    subject_file_map(path) yields a filemap object for the given HCP subject path.
    '''
    if name is None:
        if is_pseudo_path(path):
            pmod = path._path_data['pathmod']
            name = pmod.split(path.source_path)[-1]
        else: name = os.path.split(path)[-1]
    return file_map(path, hcp_filemap_instructions,
                    data_hierarchy=hcp_filemap_data_hierarchy,
                    path_parameters={'id':name})

# We adapt a section of the FreeSurfer spec to the HCP because the HCP directories include a
# mostly-complete FreeSurfer directory
def _hcp_adapt_fsinstructions(instr):
    instr = copy.copy(instr)
    if pimms.is_list(instr):
        res = []
        for (fnm,desc) in zip(instr[0::2], instr[1::2]):
            desc = _hcp_adapt_fsinstructions(desc)
            desc = (desc,) if pimms.is_map(desc) else desc
            if len(desc) == 0: continue
            res.append(fnm)
            res.append(desc[0] if len(desc) == 1 else desc)
        return res
    elif pimms.is_map(instr):
        if   'image'     in instr: instr['image'] = 'freesurfer_' + instr['image']
        elif 'raw_image' in instr: instr['raw_image'] = 'freesurfer_' + instr['raw_image']
        elif 'surface'   in instr: instr['surface'] = 'freesurfer_' + instr['surface']
        if   'hemi'      in instr:
            h = instr['hemi']
            if h.endswith('x'): return ()
            instr['hemi'] = h + '_native_MSMAll'
            instr = (instr, copy.copy(instr), copy.copy(instr))
            instr[1]['hemi'] = h + '_native_MSMSulc'
            instr[2]['hemi'] = h + '_native_FS'
        # others: label/annot/property -- these can stay the same
        return instr
    elif pimms.is_tuple(instr):
        tup = [_hcp_adapt_fsinstructions(u) for u in instr]
        tup = tuple([v for u in tup for v in (u if pimms.is_tuple(u) else [u])])
        return tup
    else: raise ValueError('Unrecognized instruction type: %s [%s]' % (type(instr), instr))
hcp_adapted_freesurfer_subject_filemap_instructions = []
# setup the freesurfer subject 
for (dname,contents) in zip(freesurfer_subject_filemap_instructions[0::2],
                            freesurfer_subject_filemap_instructions[1::2]):
    if dname == 'xhemi': continue
    hcp_adapted_freesurfer_subject_filemap_instructions.append(dname)
    hcp_adapted_freesurfer_subject_filemap_instructions.append(_hcp_adapt_fsinstructions(contents))
def _filt_coords(msh): return msh.coordinates
def _filt_faces(msh): return msh.tess.faces

def _surf(name, hemibase, suffix=['_MSMAll', '_MSMSulc', '_FS'], tess=False, format='gifti'):
    res = tuple([{'surface':name, 'hemi':(hemibase+suf), 'filt':_filt_coords, 'format':format}
                 for suf in (suffix if pimms.is_vector(suffix, str) else [suffix])])
    if tess:
        rst = tuple([{'tess':name, 'hemi':(hemibase+suf), 'filt':_filt_faces, 'format':format}
                     for suf in (suffix if pimms.is_vector(suffix, str) else [suffix])])
        res = res + rst
    return res
def _surf_fs(name, hemibase, tess=False, format='gifti'):
    return _surf(name, hemibase, suffix=['_FS'], tess=tess, format=format)
def _surf_msmsulc(name, hemibase, tess=False, format='gifti'):
    return _surf(name, hemibase, suffix=['_MSMSulc'], tess=tess, format=format)
def _surf_fssulc(name, hemibase, tess=False, format='gifti'):
    return _surf(name, hemibase, suffix=['_MSMSulc', '_FS'], tess=tess, format=format)
def _surf_msmall(name, hemibase, tess=False, format='gifti'):
    return _surf(name, hemibase, suffix=['_MSMAll'], tess=tess, format=format)
def _reg(name, hemibase, suffix=['_MSMAll', '_MSMSulc', '_FS'], format='gifti'):
    return tuple([{'registration':name, 'hemi':(hemibase+suf), 'filt':_filt_coords, 'format':format}
                  for suf in (suffix if pimms.is_vector(suffix, str) else [suffix])])
def _reg_fs(name, hemibase, format='gifti'):
    return _reg(name, hemibase, suffix=['_FS'], format=format)
def _reg_msmsulc(name, hemibase, format='gifti'):
    return _reg(name, hemibase, suffix=['_MSMSulc'], format=format)
def _reg_fssulc(name, hemibase, format='gifti'):
    return _reg(name, hemibase, suffix=['_FS', '_MSMSulc'], format=format)
def _reg_msmall(name, hemibase, format='gifti'):
    return _reg(name, hemibase, suffix=['_MSMAll'], format=format)
def _nofilt(x): return x
def _tobool(x): return np.asarray(x).astype('bool')
def _prop(name, hemibase, suffix=['_MSMAll','_MSMSulc', '_FS'],
          format='gifti', key='property', filt=None):
    tup = tuple([{key:name, 'hemi':(hemibase+suf), 'format':format}
                 for suf in (suffix if pimms.is_vector(suffix, str) else [suffix])])
    if format == 'gifti':
        for u in tup: u['filt'] = gifti_to_array
    elif format == 'cifti':
        for u in tup: u['filt'] = curry(lambda h,cii: cifti_extract(cii, h), u['hemi'])
    if key == 'label':
        for u in tup:
            u['filt'] = curry(lambda f,x: _tobool(f(x)), u['filt']) if 'filt' in u else _tobool
    if filt is not None and filt is not False:
        for u in tup:
            u['filt'] = curry(lambda ff,f,x: ff(f(x)), filt, u['filt']) if 'filt' in u else filt
    return tup
def _prop_fs(name, hemibase, format='gifti', key='property', filt=None):
    return _prop(name, hemibase, suffix=['_FS'], format=format, key=key)
def _prop_msmsulc(name, hemibase, format='gifti', key='property', filt=None):
    return _prop(name, hemibase, suffix=['_MSMSulc'], format=format, key=key)
def _prop_fssulc(name, hemibase, format='gifti', key='property', filt=None):
    return _prop(name, hemibase, suffix=['_FS', '_MSMSulc'], format=format, key=key)
def _prop_msmall(name, hemibase, format='gifti', key='property', filt=None):
    return _prop(name, hemibase, suffix=['_MSMAll'], format=format, key=key)
def _prop_cifti(name, hemires, filt=None):
    return (_prop_msmall(name, 'lh_LR%dk' % hemires, format='cifti', filt=filt) + 
            _prop_msmall(name, 'rh_LR%dk' % hemires, format='cifti', filt=filt))
def _load_imm_img(flnm, u):
    import neuropythy.mri as mri, neuropythy.io as nyio
    im = nyio.load(flnm)
    try: im.dataobj.flags['WRITEABLE'] = False
    except Exception:
        d = np.asanyarray(im.dataobj)
        d.flags['WRITEABLE'] = False
        im = type(im)(d, im.affine, im.header)
    return im

hcp_filemap_data_hierarchy = [['image'], ['raw_image'], ['flatmap'],
                              ['hemi', 'surface'], ['hemi', 'tess'],
                              ['hemi', 'registration'],
                              ['hemi', 'property'],
                              ['hemi', 'label'], ['hemi', 'alt_label'],
                              ['hemi', 'weight'], ['hemi', 'alt_weight'],
                              ['hemi', 'annot'], ['hemi', 'alt_annot']]
hcp_filemap_instructions = [
    'T1w', [
        'BiasField_acpc_dc.nii.gz',         {'image':'bias', 'load':_load_imm_img},
        'T1wDividedByT2w.nii.gz',           {'image':'T1_to_T2_ratio_all', 'load':_load_imm_img},
        'T1wDividedByT2w_ribbon.nii.gz',    {'image':'T1_to_T2_ratio', 'load':_load_imm_img},
        'T1w_acpc_dc_restore.nii.gz',       {'image':'T1', 'load':_load_imm_img},
        'T1w_acpc_dc.nii.gz',               {'image':'T1_unrestored', 'load':_load_imm_img},
        'T1w_acpc_dc_restore_brain.nii.gz', {'image':'brain', 'load':_load_imm_img},
        'T2w_acpc_dc_restore.nii.gz',       {'image':'T2', 'load':_load_imm_img},
        'T2w_acpc_dc.nii.gz',               {'image':'T2_unrestored', 'load':_load_imm_img},
        'T2w_acpc_dc_restore_brain.nii.gz', {'image':'T2_brain', 'load':_load_imm_img},
        'aparc+aseg.nii.gz',                {'image':'Desikan06_parcellation', 'load':_load_imm_img},
        'aparc.a2009s+aseg.nii.gz',         ({'image':'parcellation', 'load':_load_imm_img},
                                             {'image':'Destrieux09_parcellation', 'load':_load_imm_img}),
        'brainmask_fs.nii.gz',              {'image':'masked_brain', 'load':_load_imm_img},
        'ribbon.nii.gz',                    {'image':'ribbon', 'load':_load_imm_img},
        'wmparc.nii.gz',                    {'image':'white_parcellation', 'load':_load_imm_img},
        '{id}', hcp_adapted_freesurfer_subject_filemap_instructions,
        'Native', [
            '{id}.L.white.native.surf.gii',         _surf('white', 'lh_native', tess=True),
            '{id}.L.midthickness.native.surf.gii',  _surf('midgray', 'lh_native'),
            '{id}.L.pial.native.surf.gii',          _surf('pial', 'lh_native'),
            '{id}.L.inflated.native.surf.gii',      _surf('inflated', 'lh_native'),
            '{id}.L.very_inflated.native.surf.gii', _surf('very_inflated', 'lh_native'),
            '{id}.R.white.native.surf.gii',         _surf('white', 'rh_native', tess=True),
            '{id}.R.midthickness.native.surf.gii',  _surf('midgray', 'rh_native'),
            '{id}.R.pial.native.surf.gii',          _surf('pial', 'rh_native'),
            '{id}.R.inflated.native.surf.gii',      _surf('inflated', 'rh_native'),
            '{id}.R.very_inflated.native.surf.gii', _surf('very_inflated', 'rh_native')],
        'fsaverage_LR32k', [
            '{id}.L.inflated.32k_fs_LR.surf.gii',      _surf_fssulc('inflated',      'lh_nat32k'),
            '{id}.L.midthickness.32k_fs_LR.surf.gii',  _surf_fssulc('migray',        'lh_nat32k'),
            '{id}.L.pial.32k_fs_LR.surf.gii',          _surf_fssulc('pial',          'lh_nat32k'),
            '{id}.L.very_inflated.32k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'lh_nat32k'),
            '{id}.L.white.32k_fs_LR.surf.gii',         _surf_fssulc('white',         'lh_nat32k',
                                                                    tess=True),
            '{id}.R.inflated.32k_fs_LR.surf.gii',      _surf_fssulc('inflated',      'rh_nat32k'),
            '{id}.R.midthickness.32k_fs_LR.surf.gii',  _surf_fssulc('migray',        'rh_nat32k'),
            '{id}.R.pial.32k_fs_LR.surf.gii',          _surf_fssulc('pial',          'rh_nat32k'),
            '{id}.R.very_inflated.32k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'rh_nat32k'),
            '{id}.R.white.32k_fs_LR.surf.gii',         _surf_fssulc('white',         'rh_nat32k',
                                                                     tess=True),
            
            '{id}.L.inflated_MSMAll.32k_fs_LR.surf.gii',      _surf_msmall('inflated', 'lh_nat32k'),
            '{id}.L.midthickness_MSMAll.32k_fs_LR.surf.gii',  _surf_msmall('midgray', 'lh_nat32k'),
            '{id}.L.pial_MSMAll.32k_fs_LR.surf.gii',          _surf_msmall('pial', 'lh_nat32k'),
            '{id}.L.very_inflated_MSMAll.32k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                           'lh_nat32k'),
            '{id}.L.white_MSMAll.32k_fs_LR.surf.gii',         _surf_msmall('white', 'lh_nat32k',
                                                                           tess=True),
            '{id}.R.inflated_MSMAll.32k_fs_LR.surf.gii',      _surf_msmall('inflated', 'rh_nat32k'),
            '{id}.R.midthickness_MSMAll.32k_fs_LR.surf.gii',  _surf_msmall('midgray', 'rh_nat32k'),
            '{id}.R.pial_MSMAll.32k_fs_LR.surf.gii',          _surf_msmall('pial', 'rh_nat32k'),
            '{id}.R.very_inflated_MSMAll.32k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                           'rh_nat32k'),
            '{id}.R.white_MSMAll.32k_fs_LR.surf.gii',         _surf_msmall('white', 'rh_nat32k',
                                                                           tess=True)],
        'fsaverage_LR59k', [
            '{id}.L.inflated.59k_fs_LR.surf.gii',      _surf_fssulc('inflated',      'lh_nat59k'),
            '{id}.L.midthickness.59k_fs_LR.surf.gii',  _surf_fssulc('midgray',       'lh_nat59k'),
            '{id}.L.pial.59k_fs_LR.surf.gii',          _surf_fssulc('pial',          'lh_nat59k'),
            '{id}.L.very_inflated.59k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'lh_nat59k'),
            '{id}.L.white.59k_fs_LR.surf.gii',         _surf_fssulc('white',         'lh_nat59k',
                                                                    tess=True),
            '{id}.R.inflated.59k_fs_LR.surf.gii',      _surf_fssulc('inflated',      'rh_nat59k'),
            '{id}.R.midthickness.59k_fs_LR.surf.gii',  _surf_fssulc('midgray',       'rh_nat59k'),
            '{id}.R.pial.59k_fs_LR.surf.gii',          _surf_fssulc('pial',          'rh_nat59k'),
            '{id}.R.very_inflated.59k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'rh_nat59k'),
            '{id}.R.white.59k_fs_LR.surf.gii',         _surf_fssulc('white',         'rh_nat59k',
                                                                    tess=True),
            
            '{id}.L.inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii',      _surf_msmall('inflated',
                                                                                 'lh_nat59k'),
            '{id}.L.midthickness_1.6mm_MSMAll.59k_fs_LR.surf.gii',  _surf_msmall('midgray',
                                                                                 'lh_nat59k'),
            '{id}.L.pial_1.6mm_MSMAll.59k_fs_LR.surf.gii',          _surf_msmall('pial',
                                                                                 'lh_nat59k'),
            '{id}.L.very_inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                                 'lh_nat59k'),
            '{id}.L.white_1.6mm_MSMAll.59k_fs_LR.surf.gii',         _surf_msmall('white',
                                                                                 'lh_nat59k',
                                                                                 tess=True),
            '{id}.R.inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii',      _surf_msmall('inflated',
                                                                                 'rh_nat59k'),
            '{id}.R.midthickness_1.6mm_MSMAll.59k_fs_LR.surf.gii',  _surf_msmall('midgray',
                                                                                 'rh_nat59k'),
            '{id}.R.pial_1.6mm_MSMAll.59k_fs_LR.surf.gii',          _surf_msmall('pial',
                                                                                 'rh_nat59k'),
            '{id}.R.very_inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                                 'rh_nat59k'),
            '{id}.R.white_1.6mm_MSMAll.59k_fs_LR.surf.gii',         _surf_msmall('white',
                                                                                 'rh_nat59k',
                                                                                 tess=True)]],
    'MNINonLinear', [
        'BiasField.nii.gz',         {'image':'bias_warped', 'load':_load_imm_img},
        'T1w_restore.nii.gz',       {'image':'T1_warped', 'load':_load_imm_img},
        'T1w.nii.gz',               {'image':'T1_warped_unrestored', 'load':_load_imm_img},
        'T1w_restore_brain.nii.gz', {'image':'brain_warped', 'load':_load_imm_img},
        'T2w_restore.nii.gz',       {'image':'T2_warped', 'load':_load_imm_img},
        'T2w.nii.gz',               {'image':'T2_warped_unrestored', 'load':_load_imm_img},
        'T2w_restore_brain.nii.gz', {'image':'T2_brain_warped', 'load':_load_imm_img},
        'aparc+aseg.nii.gz',        {'image':'Desikan06_parcellation_warped', 'load':_load_imm_img},
        'aparc.a2009s+aseg.nii.gz', ({'image':'parcellation_warped', 'load':_load_imm_img},
                                     {'image':'Destrieux09_parcellation_warped', 'load':_load_imm_img}),
        'brainmask_fs.nii.gz',      {'image':'masked_brain_warped', 'load':_load_imm_img},
        'ribbon.nii.gz',            {'image':'ribbon_warped', 'load':_load_imm_img},
        'wmparc.nii.gz',            {'image':'white_parcellation_warped', 'load':_load_imm_img},
        '{id}.L.ArealDistortion_FS.164k_fs_LR.shape.gii',      _prop_fs('areal_distortion',
                                                                        'lh_LR164k'),
        '{id}.L.ArealDistortion_MSMSulc.164k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                             'lh_LR164k'),
        '{id}.L.ArealDistortion_MSMAll.164k_fs_LR.shape.gii',  _prop_msmall('areal_distortion',
                                                                            'lh_LR164k'),
        '{id}.L.MyelinMap.164k_fs_LR.func.gii',                _prop_fssulc('myelin_uncorrected',
                                                                            'lh_LR164k'),
        '{id}.L.MyelinMap_BC.164k_fs_LR.func.gii',             _prop_fssulc('myelin',
                                                                            'lh_LR164k'),
        '{id}.L.SmoothedMyelinMap.164k_fs_LR.func.gii',  _prop_fssulc('myelin_smooth_uncorrected',
                                                                      'lh_LR164k'),
        '{id}.L.SmoothedMyelinMap_BC.164k_fs_LR.func.gii',     _prop_fssulc('myelin_smooth',
                                                                            'lh_LR164k'),
        '{id}.L.RefMyelinMap.164k_fs_LR.func.gii',             _prop_fssulc('myelin_ref',
                                                                            'lh_LR164k'),
        '{id}.L.BA.164k_fs_LR.label.gii',                      _prop_msmall('brodmann_area',
                                                                            'lh_LR164k'),
        '{id}.L.aparc.164k_fs_LR.label.gii',         _prop('Desikan06_parcellation', 'lh_LR164k'),
        '{id}.L.aparc.a2009s.164k_fs_LR.label.gii',  (_prop('parcellation', 'lh_LH164k') +
                                                      _prop('Destrieux09_parcellation',
                                                            'lh_LH164k')),
        '{id}.L.atlasroi.164k_fs_LR.shape.gii',      _prop('atlas', 'lh_LR164k', key='label'),
        '{id}.L.curvature.164k_fs_LR.shape.gii',     _prop_fssulc('curvature', 'lh_LR164k',
                                                                  filt=lambda c:-c),
        '{id}.L.sulc.164k_fs_LR.shape.gii',          _prop_fssulc('convexity', 'lh_LR164k'),
        '{id}.L.corrThickness.164k_fs_LR.shape.gii', _prop_fssulc('thickness', 'lh_LR164k'),
        '{id}.L.thickness.164k_fs_LR.shape.gii',     _prop_fssulc('thickness_uncorrected',
                                                                   'lh_LR164k'),
        '{id}.L.white.164k_fs_LR.surf.gii',         _surf_fssulc('white', 'lh_LR164k', tess=True),
        '{id}.L.midthickness.164k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'lh_LR164k'),
        '{id}.L.pial.164k_fs_LR.surf.gii',          _surf_fssulc('pial', 'lh_LR164k'),
        '{id}.L.inflated.164k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'lh_LR164k'),
        '{id}.L.very_inflated.164k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'lh_LR164k'),
        '{id}.L.midthickness_MSMAll.164k_fs_LR.surf.gii',  _surf_msmall('midgray', 'lh_LR164k'),
        '{id}.L.pial_MSMAll.164k_fs_LR.surf.gii',          _surf_msmall('pial', 'lh_LR164k'), 
        '{id}.L.inflated_MSMAll.164k_fs_LR.surf.gii',      _surf_msmall('inflated', 'lh_LR164k'),
        '{id}.L.very_inflated_MSMAll.164k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                        'lh_LR164k'),
        '{id}.L.white_MSMAll.164k_fs_LR.surf.gii',          _surf_msmall('white', 'lh_LR164k',
                                                                         tess=True),
        '{id}.L.sphere.164k_fs_LR.surf.gii',               _reg('fs_LR', 'lh_LR164k'),
        '{id}.L.flat.164k_fs_LR.surf.gii', ({'flatmap':'lh_LR164k_FS',      'format':'gifti'},
                                            {'flatmap':'lh_LR164k_MSMSulc', 'format':'gifti'},
                                            {'flatmap':'lh_LR164k_MSMAll',  'format':'gifti'}),
        '{id}.R.ArealDistortion_FS.164k_fs_LR.shape.gii',      _prop_fs('areal_distortion',
                                                                        'rh_LR164k'),
        '{id}.R.ArealDistortion_MSMSulc.164k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                             'rh_LR164k'),
        '{id}.R.ArealDistortion_MSMAll.164k_fs_LR.shape.gii',  _prop_msmall('areal_distortion',
                                                                            'rh_LR164k'),
        '{id}.R.MyelinMap.164k_fs_LR.func.gii',                _prop_fssulc('myelin_uncorrected',
                                                                            'rh_LR164k'),
        '{id}.R.MyelinMap_BC.164k_fs_LR.func.gii',             _prop_fssulc('myelin', 'rh_LR164k'),
        '{id}.R.SmoothedMyelinMap.164k_fs_LR.func.gii',   _prop_fssulc('myelin_smooth_uncorrected',
                                                                       'rh_LR164k'),
        '{id}.R.SmoothedMyelinMap_BC.164k_fs_LR.func.gii',     _prop_fssulc('myelin_smooth',
                                                                            'rh_LR164k'),
        '{id}.R.RefMyelinMap.164k_fs_LR.func.gii',             _prop_fssulc('myelin_ref',
                                                                            'rh_LR164k'),
        '{id}.R.BA.164k_fs_LR.label.gii',                      _prop_msmall('brodmann_area',
                                                                            'rh_LR164k'),
        '{id}.R.aparc.164k_fs_LR.label.gii',         _prop('Desikan06_parcellation',
                                                           'rh_LR164k'),
        '{id}.R.aparc.a2009s.164k_fs_LR.label.gii',  (_prop('parcellation', 'rh_LH164k') +
                                                      _prop('Destrieux09_parcellation',
                                                            'rh_LH164k')),
        '{id}.R.atlasroi.164k_fs_LR.shape.gii',      _prop('atlas', 'rh_LR164k', key='label'),
        '{id}.R.curvature.164k_fs_LR.shape.gii',     _prop_fssulc('curvature', 'rh_LR164k',
                                                                  filt=lambda c:-c),
        '{id}.R.sulc.164k_fs_LR.shape.gii',          _prop_fssulc('convexity', 'rh_LR164k'),
        '{id}.R.corrThickness.164k_fs_LR.shape.gii', _prop_fssulc('thickness', 'rh_LR164k'),
        '{id}.R.thickness.164k_fs_LR.shape.gii',     _prop_fssulc('thickness_uncorrected',
                                                                  'rh_LR164k'),
        '{id}.R.white.164k_fs_LR.surf.gii',         _surf_fssulc('white', 'rh_LR164k', tess=True),
        '{id}.R.midthickness.164k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'rh_LR164k'),
        '{id}.R.pial.164k_fs_LR.surf.gii',          _surf_fssulc('pial', 'rh_LR164k'),
        '{id}.R.inflated.164k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'rh_LR164k'),
        '{id}.R.very_inflated.164k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'rh_LR164k'),
        '{id}.R.midthickness_MSMAll.164k_fs_LR.surf.gii',  _surf_msmall('midgray', 'rh_LR164k'),
        '{id}.R.pial_MSMAll.164k_fs_LR.surf.gii',          _surf_msmall('pial', 'rh_LR164k'), 
        '{id}.R.inflated_MSMAll.164k_fs_LR.surf.gii',      _surf_msmall('inflated', 'rh_LR164k'),
        '{id}.R.very_inflated_MSMAll.164k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                        'lh_LR164k'),
        '{id}.R.white_MSMAll.164k_fs_LR.surf.gii',          _surf_msmall('white', 'rh_LR164k',
                                                                         tess=True),
        '{id}.R.sphere.164k_fs_LR.surf.gii',               _reg('fs_LR', 'rh_LR164k'),
        '{id}.R.flat.164k_fs_LR.surf.gii', ({'flatmap':'rh_LR164k_FS',      'format':'gifti'},
                                            {'flatmap':'rh_LR164k_MSMSulc', 'format':'gifti'},
                                            {'flatmap':'rh_LR164k_MSMAll',  'format':'gifti'}),
        '{id}.ArealDistortion_MSMAll.164k_fs_LR.dscalar.nii',      _prop_cifti('areal_distortion',
                                                                               164),
        '{id}.MyelinMap_BC_MSMAll.164k_fs_LR.dscalar.nii',         _prop_cifti('myelin', 164),
        '{id}.SmoothedMyelinMap_BC_MSMAll.164k_fs_LR.dscalar.nii', _prop_cifti('myelin_smooth',
                                                                               164),
        '{id}.curvature_MSMAll.164k_fs_LR.dscalar.nii',            _prop_cifti('curvature', 164,
                                                                               filt=lambda c:-c),
        '{id}.sulc.164k_fs_LR.dscalar.nii',                        _prop_cifti('convexity', 164),
        '{id}.corrThickness.164k_fs_LR.dscalar.nii',               _prop_cifti('thickness', 164),
        '{id}.thickness.164k_fs_LR.dscalar.nii', _prop_cifti('thickness_uncorrected', 164),
        'Native', [
            '{id}.L.ArealDistortion_FS.native.shape.gii', _prop_fs('areal_distortion', 'lh_native'),
            '{id}.L.ArealDistortion_MSMSulc.native.shape.gii', _prop_msmsulc('areal_distortion',
                                                                             'lh_native'),
            '{id}.L.ArealDistortion_MSMAll.native.shape.gii',  _prop_msmall('areal_distortion',
                                                                            'lh_native'),
            '{id}.L.MyelinMap.native.func.gii', _prop_fssulc('myelin_uncorrected', 'lh_native'),
            '{id}.L.MyelinMap_BC.native.func.gii', _prop_fssulc('myelin', 'lh_native'),
            '{id}.L.SmoothedMyelinMap.native.func.gii', _prop_fssulc('myelin_smooth_uncorrected',
                                                                     'lh_native'),
            '{id}.L.SmoothedMyelinMap_BC.native.func.gii', _prop_fssulc('myelin_smooth',
                                                                        'lh_native'),
            '{id}.L.RefMyelinMap.native.func.gii', _prop_fssulc('myelin_ref', 'lh_native'),
            '{id}.L.BA.native.label.gii', _prop('brodmann_area', 'lh_native'),
            '{id}.L.aparc.native.label.gii',  _prop('Deskian06_parcellation', 'lh_native'),
            '{id}.L.aparc.a2009s.native.label.gii', (_prop('Destrieux09_parcellation', 'lh_native')
                                                     + _prop('parcellation', 'lh_native')),
            '{id}.L.atlasroi.native.shape.gii', _prop('atlas', 'lh_native', key='label'),
            '{id}.L.curvature.native.shape.gii', _prop('curvature', 'lh_native', filt=lambda c:-c),
            '{id}.L.sulc.native.shape.gii', _prop('convexity', 'lh_native'),
            '{id}.L.corrThickness.native.shape.gii', _prop('thickness', 'lh_native'),
            '{id}.L.thickness.native.shape.gii', _prop('thickness_uncorrected', 'lh_native'),
            '{id}.L.roi.native.shape.gii', _prop('roi', 'lh_native', filt=_tobool),
            '{id}.L.sphere.native.surf.gii',     _reg('native', 'lh_native'),
            '{id}.L.sphere.reg.native.surf.gii', _reg('fsaverage', 'lh_native'),
            '{id}.L.sphere.reg.reg_LR.native.surf.gii', _reg_fs('fs_LR', 'lh_native'),
            '{id}.L.sphere.MSMAll.native.surf.gii', _reg_msmall('fs_LR', 'lh_native'),
            '{id}.L.sphere.MSMSulc.native.surf.gii',_reg_msmsulc('fs_LR', 'lh_native'),
            '{id}.R.ArealDistortion_FS.native.shape.gii', _prop_fs('areal_distortion', 'rh_native'),
            '{id}.R.ArealDistortion_MSMSulc.native.shape.gii', _prop_msmsulc('areal_distortion',
                                                                             'rh_native'),
            '{id}.R.ArealDistortion_MSMAll.native.shape.gii',  _prop_msmall('areal_distortion',
                                                                            'rh_native'),
            '{id}.R.MyelinMap.native.func.gii', _prop_fssulc('myelin_uncorrected', 'rh_native'),
            '{id}.R.MyelinMap_BC.native.func.gii', _prop_fssulc('myelin', 'rh_native'),
            '{id}.R.SmoothedMyelinMap.native.func.gii', _prop_fssulc('myelin_smooth_uncorrected',
                                                                      'rh_native'),
            '{id}.R.SmoothedMyelinMap_BC.native.func.gii', _prop_fssulc('myelin_smooth',
                                                                         'rh_native'),
            '{id}.R.RefMyelinMap.native.func.gii', _prop_fssulc('myelin_ref', 'rh_native'),
            '{id}.R.BA.native.label.gii', _prop('brodmann_area', 'rh_native'),
            '{id}.R.aparc.native.label.gii',  _prop('Deskian06_parcellation', 'rh_native'),
            '{id}.R.aparc.a2009s.native.label.gii', (_prop('Destrieux09_parcellation', 'rh_native')
                                                    + _prop('parcellation', 'rh_native')),
            '{id}.R.atlasroi.native.shape.gii', _prop('atlas', 'rh_native', key='label'),
            '{id}.R.curvature.native.shape.gii', _prop('curvature', 'rh_native', filt=lambda c:-c),
            '{id}.R.sulc.native.shape.gii', _prop('convexity', 'rh_native'),
            '{id}.R.corrThickness.native.shape.gii', _prop('thickness', 'rh_native'),
            '{id}.R.thickness.native.shape.gii', _prop('thickness_uncorrected', 'rh_native'),
            '{id}.R.roi.native.shape.gii', _prop('roi', 'rh_native', key='label'),
            '{id}.R.sphere.native.surf.gii',     _reg('native', 'rh_native'),
            '{id}.R.sphere.reg.native.surf.gii', _reg('fsaverage', 'rh_native'),
            '{id}.R.sphere.reg.reg_LR.native.surf.gii', _reg_fs('fs_LR', 'rh_native'),
            '{id}.R.sphere.MSMAll.native.surf.gii', _reg_msmall('fs_LR', 'rh_native'),
            '{id}.R.sphere.MSMSulc.native.surf.gii',_reg_msmsulc('fs_LR', 'rh_native')],
        'fsaverage_LR59k', [
            '{id}.L.BA.59k_fs_LR.label.gii', _prop('brodmann_area', 'lh_LR59k'),
            '{id}.L.aparc.59k_fs_LR.label.gii', _prop('Desikan06_parcellation', 'lh_LR59k'),
            '{id}.L.aparc.a2009s.59k_fs_LR.label.gii', (_prop('Destrieux09_parcellation',
                                                              'lh_LR59k') + 
                                                        _prop('parcellation', 'lh_LR59k')),
            '{id}.L.ArealDistortion_FS.59k_fs_LR.shape.gii', _prop_fs('areal_distortion',
                                                                      'lh_LR59k'),
            '{id}.L.ArealDistortion_MSMSulc.59k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                                'lh_LR59k'),
            '{id}.L.ArealDistortion_MSMAll.59k_fs_LR.shape.gii',  _prop_msmall('areal_distortion',
                                                                               'lh_LR59k'),
            '{id}.L.MyelinMap.59k_fs_LR.func.gii', _prop_fssulc('myelin_uncorrected', 'lh_LR59k'),
            '{id}.L.MyelinMap_BC.59k_fs_LR.func.gii', _prop_fssulc('myelin', 'lh_LR59k'),
            '{id}.L.SmoothedMyelinMap.59k_fs_LR.func.gii',_prop_fssulc('myelin_smooth_uncorrected',
                                                                       'lh_LR59k'),
            '{id}.L.SmoothedMyelinMap_BC.59k_fs_LR.func.gii', _prop_fssulc('myelin_smooth',
                                                                           'lh_LR59k'),
            '{id}.L.RefMyelinMap.59k_fs_LR.func.gii', _prop_fssulc('myelin_ref', 'lh_LR59k'),
            '{id}.L.atlasroi.59k_fs_LR.shape.gii', _prop('atlas', 'lh_LR59k', key='label'),
            '{id}.L.curvature.59k_fs_LR.shape.gii', _prop('curvature', 'lh_LR59k',
                                                          filt=lambda c:-c),
            '{id}.L.sulc.59k_fs_LR.shape.gii', _prop('convexity', 'lh_LR59k'),
            '{id}.L.corrThickness.59k_fs_LR.shape.gii', _prop('thickness', 'lh_LR59k'),
            '{id}.L.thickness.59k_fs_LR.shape.gii', _prop('thickness_uncorrected', 'lh_LR59k'),
            '{id}.L.midthickness.59k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'lh_LR59k'),
            '{id}.L.pial.59k_fs_LR.surf.gii',          _surf_fssulc('pial', 'lh_LR59k'),
            '{id}.L.inflated.59k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'lh_LR59k'),
            '{id}.L.very_inflated.59k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'lh_LR59k'),
            '{id}.L.white.59k_fs_LR.surf.gii',         _surf_fssulc('white', 'lh_LR59k',
                                                                     tess=True),
            '{id}.L.white_1.6mm_MSMAll.59k_fs_LR.surf.gii',         _surf_msmall('white',
                                                                                 'lh_LR59k',
                                                                                 tess=True),
            '{id}.L.midthickness_1.6mm_MSMAll.59k_fs_LR.surf.gii',  _surf_msmall('midgray',
                                                                                 'lh_LR59k'),
            '{id}.L.pial_1.6mm_MSMAll.59k_fs_LR.surf.gii',          _surf_msmall('pial',
                                                                                 'lh_LR59k'),
            '{id}.L.inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii',      _surf_msmall('inflated',
                                                                                 'lh_LR59k'),
            '{id}.L.very_inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                                 'lh_LR59k'),
            '{id}.L.sphere.59k_fs_LR.surf.gii', _reg('fs_LR', 'lh_LR59k'),
            '{id}.L.flat.59k_fs_LR.surf.gii', ({'flatmap':'lh_LR59k_FS',      'format':'gifti'},
                                               {'flatmap':'lh_LR59k_MSMSulc', 'format':'gifti'},
                                               {'flatmap':'lh_LR59k_MSMAll',  'format':'gifti'}),
            '{id}.R.BA.59k_fs_LR.label.gii', _prop('brodmann_area', 'rh_LR59k'),
            '{id}.R.aparc.59k_fs_LR.label.gii', _prop('Desikan06_parcellation', 'rh_LR59k'),
            '{id}.R.aparc.a2009s.59k_fs_LR.label.gii', (_prop('Destrieux09_parcellation',
                                                              'rh_LR59k') + 
                                                        _prop('parcellation', 'rh_LR59k')),
            '{id}.R.ArealDistortion_FS.59k_fs_LR.shape.gii', _prop_fs('areal_distortion',
                                                                      'rh_LR59k'),
            '{id}.R.ArealDistortion_MSMSulc.59k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                                'rh_LR59k'),
            '{id}.R.ArealDistortion_MSMAll.59k_fs_LR.shape.gii', _prop_msmall('areal_distortion',
                                                                              'rh_LR59k'),
            '{id}.R.MyelinMap.59k_fs_LR.func.gii', _prop_fssulc('myelin_uncorrected', 'rh_LR59k'),
            '{id}.R.MyelinMap_BC.59k_fs_LR.func.gii', _prop_fssulc('myelin', 'rh_LR59k'),
            '{id}.R.SmoothedMyelinMap.59k_fs_LR.func.gii',_prop_fssulc('myelin_smooth_uncorrected',
                                                                       'rh_LR59k'),
            '{id}.R.SmoothedMyelinMap_BC.59k_fs_LR.func.gii', _prop_fssulc('myelin_smooth',
                                                                           'rh_LR59k'),
            '{id}.R.RefMyelinMap.59k_fs_LR.func.gii', _prop_fssulc('myelin_ref', 'rh_LR59k'),
            '{id}.R.atlasroi.59k_fs_LR.shape.gii', _prop('atlas', 'rh_LR59k', key='label'),
            '{id}.R.curvature.59k_fs_LR.shape.gii', _prop('curvature', 'rh_LR59k',
                                                          filt=lambda c:-c),
            '{id}.R.sulc.59k_fs_LR.shape.gii', _prop_fssulc('convexity', 'rh_LR59k'),
            '{id}.R.corrThickness.59k_fs_LR.shape.gii', _prop_fssulc('thickness', 'rh_LR59k'),
            '{id}.R.thickness.59k_fs_LR.shape.gii', _prop_fssulc('thickness_uncorrected',
                                                                 'rh_LR59k'),
            '{id}.R.midthickness.59k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'rh_LR59k'),
            '{id}.R.pial.59k_fs_LR.surf.gii',          _surf_fssulc('pial', 'rh_LR59k'),
            '{id}.R.inflated.59k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'rh_LR59k'),
            '{id}.R.very_inflated.59k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'rh_LR59k'),
            '{id}.R.white.59k_fs_LR.surf.gii',         _surf_fssulc('white', 'rh_LR59k',
                                                                    tess=True),
            '{id}.R.white_1.6mm_MSMAll.59k_fs_LR.surf.gii',         _surf_msmall('white',
                                                                                 'rh_LR59k',
                                                                                 tess=True),
            '{id}.R.midthickness_1.6mm_MSMAll.59k_fs_LR.surf.gii',  _surf_msmall('midgray',
                                                                                 'rh_LR59k'),
            '{id}.R.pial_1.6mm_MSMAll.59k_fs_LR.surf.gii',          _surf_msmall('pial',
                                                                                 'rh_LR59k'),
            '{id}.R.inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii',      _surf_msmall('inflated',
                                                                                 'rh_LR59k'),
            '{id}.R.very_inflated_1.6mm_MSMAll.59k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                                 'rh_LR59k'),
            '{id}.R.flat.59k_fs_LR.surf.gii', ({'flatmap':'rh_LR59k_FS',      'format':'gifti'},
                                               {'flatmap':'rh_LR59k_MSMSulc', 'format':'gifti'},
                                               {'flatmap':'rh_LR59k_MSMAll',  'format':'gifti'}),
            '{id}.R.sphere.59k_fs_LR.surf.gii', _reg('fs_LR', 'rh_LR59k'),
            '{id}.MyelinMap_1.6mm_MSMAll.59k_fs_LR.dscalar.nii', _prop_cifti('myelin_uncorrected',
                                                                             59),
            '{id}.MyelinMap_BC_1.6mm_MSMAll.59k_fs_LR.dscalar.nii', _prop_cifti('myelin', 59),
            '{id}.SmoothedMyelinMap_BC_1.6mm_MSMAll.59k_fs_LR.dscalar.nii', _prop_cifti(
                'myelin_smooth', 59),
            '{id}.ArealDistortion_1.6mm_MSMAll.59k_fs_LR.shape.nii', _prop_cifti('areal_distortion',
                                                                                 59),
            '{id}.curvature_1.6mm_MSMAll.59k_fs_LR.shape.nii', _prop_cifti('curvature', 59,
                                                                           filt=lambda c:-c),
            '{id}.sulc_1.6mm_MSMAll.59k_fs_LR.dscalar.nii', _prop_cifti('convexity', 59),
            '{id}.thickness_1.6mm_MSMAll.59k_fs_LR.dscalar.nii',_prop_cifti('thickness_uncorrected',
                                                                            59),
            '{id}.corrThickness_1.6mm_MSMAll.59k_fs_LR.dscalar.nii', _prop_cifti('thickness', 59)],
        'fsaverage_LR32k', [
            '{id}.L.BA.32k_fs_LR.label.gii', _prop('brodmann_area', 'lh_LR32k'),
            '{id}.L.aparc.32k_fs_LR.label.gii', _prop('Desikan06_parcellation', 'lh_LR32k'),
            '{id}.L.aparc.a2009s.32k_fs_LR.label.gii', (_prop('Destrieux09_parcellation',
                                                              'lh_LR32k') + 
                                                        _prop('parcellation', 'lh_LR32k')),
            '{id}.L.ArealDistortion_FS.32k_fs_LR.shape.gii', _prop_fs('areal_distortion',
                                                                      'lh_LR32k'),
            '{id}.L.ArealDistortion_MSMSulc.32k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                                'lh_LR32k'),
            '{id}.L.ArealDistortion_MSMAll.32k_fs_LR.shape.gii', _prop_msmall('areal_distortion',
                                                                              'lh_LR32k'),
            '{id}.L.MyelinMap.32k_fs_LR.func.gii', _prop_fssulc('myelin_uncorrected', 'lh_LR32k'),
            '{id}.L.MyelinMap_BC.32k_fs_LR.func.gii', _prop_fssulc('myelin', 'lh_LR32k'),
            '{id}.L.SmoothedMyelinMap.32k_fs_LR.func.gii',_prop_fssulc('myelin_smooth_uncorrected',
                                                                       'lh_LR32k'),
            '{id}.L.SmoothedMyelinMap_BC.32k_fs_LR.func.gii', _prop_fssulc('myelin_smooth',
                                                                           'lh_LR32k'),
            '{id}.L.RefMyelinMap.32k_fs_LR.func.gii', _prop_fssulc('myelin_ref', 'lh_LR32k'),
            '{id}.L.atlasroi.32k_fs_LR.shape.gii', _prop('atlas', 'lh_LR32k', key='label'),
            '{id}.L.curvature.32k_fs_LR.shape.gii', _prop('curvature', 'lh_LR32k',
                                                          filt=lambda c:-c),
            '{id}.L.sulc.32k_fs_LR.shape.gii', _prop('convexity', 'lh_LR32k'),
            '{id}.L.corrThickness.32k_fs_LR.shape.gii', _prop('thickness', 'lh_LR32k'),
            '{id}.L.thickness.32k_fs_LR.shape.gii', _prop('thickness_uncorrected', 'lh_LR32k'),
            '{id}.L.midthickness.32k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'lh_LR32k'),
            '{id}.L.pial.32k_fs_LR.surf.gii',          _surf_fssulc('pial', 'lh_LR32k'),
            '{id}.L.inflated.32k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'lh_LR32k'),
            '{id}.L.very_inflated.32k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'lh_LR32k'),
            '{id}.L.white.32k_fs_LR.surf.gii',         _surf_fssulc('white', 'lh_LR32k',
                                                                     tess=True),
            '{id}.L.white_MSMAll.32k_fs_LR.surf.gii',         _surf_msmall('white',
                                                                                 'lh_LR32k',
                                                                                 tess=True),
            '{id}.L.midthickness_MSMAll.32k_fs_LR.surf.gii',  _surf_msmall('midgray',
                                                                                 'lh_LR32k'),
            '{id}.L.pial_MSMAll.32k_fs_LR.surf.gii',          _surf_msmall('pial',
                                                                                 'lh_LR32k'),
            '{id}.L.inflated_MSMAll.32k_fs_LR.surf.gii',      _surf_msmall('inflated',
                                                                                 'lh_LR32k'),
            '{id}.L.very_inflated_MSMAll.32k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                                 'lh_LR32k'),
            '{id}.L.sphere.32k_fs_LR.surf.gii', _reg('fs_LR', 'lh_LR32k'),
            '{id}.L.flat.32k_fs_LR.surf.gii', ({'flatmap':'lh_LR32k_FS',      'format':'gifti'},
                                               {'flatmap':'lh_LR32k_MSMSulc', 'format':'gifti'},
                                               {'flatmap':'lh_LR32k_MSMAll',  'format':'gifti'}),
            '{id}.R.BA.32k_fs_LR.label.gii', _prop('brodmann_area', 'rh_LR32k'),
            '{id}.R.aparc.32k_fs_LR.label.gii', _prop('Desikan06_parcellation', 'rh_LR32k'),
            '{id}.R.aparc.a2009s.32k_fs_LR.label.gii', (_prop('Destrieux09_parcellation',
                                                              'rh_LR32k') + 
                                                        _prop('parcellation', 'rh_LR32k')),
            '{id}.R.ArealDistortion_FS.32k_fs_LR.shape.gii', _prop_fs('areal_distortion',
                                                                      'rh_LR32k'),
            '{id}.R.ArealDistortion_MSMSulc.32k_fs_LR.shape.gii', _prop_msmsulc('areal_distortion',
                                                                                'rh_LR32k'),
            '{id}.R.ArealDistortion_MSMAll.32k_fs_LR.shape.gii',  _prop_msmall('areal_distortion',
                                                                               'rh_LR32k'),
            '{id}.R.MyelinMap.32k_fs_LR.func.gii', _prop_fssulc('myelin_uncorrected', 'rh_LR32k'),
            '{id}.R.MyelinMap_BC.32k_fs_LR.func.gii', _prop_fssulc('myelin', 'rh_LR32k'),
            '{id}.R.SmoothedMyelinMap.32k_fs_LR.func.gii',_prop_fssulc('myelin_smooth_uncorrected',
                                                                        'rh_LR32k'),
            '{id}.R.SmoothedMyelinMap_BC.32k_fs_LR.func.gii', _prop_fssulc('myelin_smooth',
                                                                            'rh_LR32k'),
            '{id}.R.RefMyelinMap.32k_fs_LR.func.gii', _prop_fssulc('myelin_ref', 'rh_LR32k'),
            '{id}.R.atlasroi.32k_fs_LR.shape.gii', _prop('atlas', 'rh_LR32k', key='label'),
            '{id}.R.curvature.32k_fs_LR.shape.gii', _prop('curvature', 'rh_LR32k',
                                                          filt=lambda c:-c),
            '{id}.R.sulc.32k_fs_LR.shape.gii', _prop_fssulc('convexity', 'rh_LR32k'),
            '{id}.R.corrThickness.32k_fs_LR.shape.gii', _prop_fssulc('thickness', 'rh_LR32k'),
            '{id}.R.thickness.32k_fs_LR.shape.gii', _prop_fssulc('thickness_uncorrected',
                                                                 'rh_LR32k'),
            '{id}.R.midthickness.32k_fs_LR.surf.gii',  _surf_fssulc('midgray', 'rh_LR32k'),
            '{id}.R.pial.32k_fs_LR.surf.gii',          _surf_fssulc('pial', 'rh_LR32k'),
            '{id}.R.inflated.32k_fs_LR.surf.gii',      _surf_fssulc('inflated', 'rh_LR32k'),
            '{id}.R.very_inflated.32k_fs_LR.surf.gii', _surf_fssulc('very_inflated', 'rh_LR32k'),
            '{id}.R.white.32k_fs_LR.surf.gii',         _surf_fssulc('white', 'rh_LR32k', tess=True),
            '{id}.R.white_MSMAll.32k_fs_LR.surf.gii',         _surf_msmall('white', 'rh_LR32k',
                                                                           tess=True),
            '{id}.R.midthickness_MSMAll.32k_fs_LR.surf.gii',  _surf_msmall('midgray', 'rh_LR32k'),
            '{id}.R.pial_MSMAll.32k_fs_LR.surf.gii',          _surf_msmall('pial', 'rh_LR32k'),
            '{id}.R.inflated_MSMAll.32k_fs_LR.surf.gii',      _surf_msmall('inflated', 'rh_LR32k'),
            '{id}.R.very_inflated_MSMAll.32k_fs_LR.surf.gii', _surf_msmall('very_inflated',
                                                                           'rh_LR32k'),
            '{id}.R.flat.32k_fs_LR.surf.gii', ({'flatmap':'rh_LR32k_FS',      'format':'gifti'},
                                               {'flatmap':'rh_LR32k_MSMSulc', 'format':'gifti'},
                                               {'flatmap':'rh_LR32k_MSMAll',  'format':'gifti'}),
            '{id}.R.sphere.32k_fs_LR.surf.gii', _reg('fs_LR', 'rh_LR32k'),
            '{id}.MyelinMap_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('myelin_uncorrected', 32),
            '{id}.MyelinMap_BC_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('myelin', 32),
            '{id}.SmoothedMyelinMap_BC_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('myelin_smooth',
                                                                                  32),
            '{id}.ArealDistortion_MSMAll.32k_fs_LR.shape.nii', _prop_cifti('areal_distortion', 32),
            '{id}.curvature_MSMAll.32k_fs_LR.shape.nii', _prop_cifti('curvature', 32,
                                                                     filt=lambda c:-c),
            '{id}.sulc_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('convexity', 32),
            '{id}.thickness_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('thickness_uncorrected', 32),
            '{id}.corrThickness_MSMAll.32k_fs_LR.dscalar.nii', _prop_cifti('thickness', 32)]]]

@nyio.importer('cifti', ('nii',))
def load_cifti(filename, to='auto'):
    '''
    load_cifti(filename) yields the cifti image referened by the given filename by using the nibabel
      load function.
    
    The optional argument to may be used to coerce the resulting data to a particular format; the
    following arguments are understood:
      * 'header' will yield just the image header
      * 'data' will yield the image's data-array
      * 'field' will yield a squeezed version of the image's data-array and will raise an error if
        the data object has more than 2 non-unitary dimensions (appropriate for loading surface
        properties stored in image files)
      * 'image' will yield the raw image object
      * 'auto' is equivalent to 'image' unless the image has no more than 2 non-unitary dimensions,
        in which case it is assumed to be a surface-field and the return value is equivalent to
        the 'field' value.
    '''
    img = nib.load(filename)
    if not isinstance(img, nib.cifti2.Cifti2Image):
        raise ValueError('given file is not a cifti image')
    to = 'auto' if to is None else to.lower()
    if   to == 'image':  return img
    elif to == 'data':   return img.dataobj
    elif to == 'header': return img.header
    elif to == 'field':
        dat = np.squeeze(np.asarray(img.dataobj))
        if len(dat.shape) > 2:
            raise ValueError('image requested as field has more than 2 non-unitary dimensions')
        return dat
    elif to in ['auto', 'automatic']:
        dims = set(np.shape(img.dataobj))
        if 1 < len(dims) < 4 and 1 in dims: return np.squeeze(np.asarray(img.dataobj))
        else:                               return img
    else:
        raise ValueError('unrecognized \'to\' argument \'%s\'' % to)



####################################################################################################
# neuropythy/io/core.py
# This file implements the load and save functions that can be used to read and write various kinds
# of neuroscience data

import numpy      as np
import pyrsistent as pyr
import nibabel    as nib
import os, six, json, gzip, pimms

# The list of import-types we understand
importers = pyr.m()
'''
neuropythy.io.core.importers is the persistent map of file-types that can be imported by neuropythy.
See also neuropythy.io.load.
'''

def guess_import_format(filename, **kwargs):
    '''
    guess_import_format(filename) attempts to guess the file format for the given filename; it does
      this guessing by looking at the file extension and using registered sniff-tests from
      importers. It will not attempt to load the file, so if the extension of the filename is
      missing, it is unlikely that this function will deduce the file-type (though load will often
      succeeed at extracting the data by trying all types exhaustively). If guess_import_format
      cannot deduce the format, it yields None.

    Note that if the filename has an extension that is recognized by neuropythy but the file itself
    is of another format, this function will never look beyond the extention in the filename;
    neither this function nor load perform that level of deduction.
    
    Keyword arguments that are passed to load should also be passed to guess_import_format.
    '''
    # first try file extension
    (_,filename) = os.path.split(filename)
    if '.' in filename:
        fnm = filename.lower()
        fmt = next((k for (k,(_,es,_,_)) in six.iteritems(importers)
                    if any(fnm.endswith('.' + e) for e in es)),
                   None)
        if fmt: return fmt
    # that didn't work; let's check the sniffers
    for (k,(_,_,sniff,_)) in six.iteritems(importers):
        try:
            if sniff(filename, **kwargs): return k
        except Exception: pass
    return None
def load(filename, format=None, **kwargs):
    '''
    load(filename) yields the data contained in the file referenced by the given filename in a
      neuropythy or neuropythy-friendly format.
    load(filename, format) specifies that the given format should be used; this should be the name
      of the importer (though a file extension that is recognized also will work).

    Additionally, functions located in load.<format> may be used; so, for example, the following
    are equivalent calls:
      load(filename, 'nifti')
      load.nifti(filename)
    In fact, the load.nifti function is just the nifti importer, so help(load.nifti) will also
    yield documentation for the nifti importer.

    Keyword options may be passed to load; these must match those accepted by the given import
    function.
    '''
    from neuropythy.util import (ObjectWithMetaData, to_pseudo_path)
    # use the pseudo-path interface for the filename:
    pdir = to_pseudo_path(filename)
    filename = pdir.local_path()
    obj = None
    if format is None:
        # Try to guess first; if this fails, we try others...
        format = guess_import_format(filename, **kwargs)
        if format is not None:
            try:
                (f,_,_,md) = importers[format]
                obj = f(filename, **kwargs)
            except Exception: format = None
        # no good guess, try them all:
        if format is None:
            for (k,(f,_,_,md)) in six.iteritems(importers):
                if obj is not None: break
                if not bool(md.get('auto', True)): continue
                try:              obj = f(filename, **kwargs)
                except Exception: pass
        if obj is None:
            raise ValueError('Could not deduce format of file %s' % filename)
    else:
        format = format.lower()
        if format not in importers:
            raise ValueError('Format \'%s\' not recognized by neuropythy' % format)
        (f,_,_,_) = importers[format]
        obj = f(filename, **kwargs)
    if isinstance(obj, ObjectWithMetaData): obj = obj.with_meta(source_filename=filename)
    return obj
def importer(name, extensions=None, sniff=None, auto=True):
    '''
    @importer(name) is a decorator that declares that the following function is an file loading
      function that should be registered with the neuropythy load function. See also the
      forget_importer function.

    Any importer function must take, as its first argument, a filename; after that it may take any
    number of keyword arguments, but no other non-keyword arguments. These keyword arguments can be
    passed to the neuropythy load function.
    
    The following options are accepted:
      * extensions (default: None) may be a string or a collection of strings that indicate possible
        file extensions for files of this type.
      * sniff (default: None) may optionally be a function f(s) that yields True when the given
        string s is a filename for a file of this type. If no sniff is given, this type can still
        be detected by running the importer and catching any raised exception.
      * auto (default: True) may optionally be set to false to specify that the importer should not
        be auto-guessed--i.e., either the sniff must yield true or the format must be explicitly
        specified for the importer to be invoked; it will never be invoked by trial and error.
    '''
    name = name.lower()
    if name in importers:
        raise ValueError('An importer for type %s already exists; see forget_importer' % name)
    if extensions is None:         extensions = ()
    elif pimms.is_str(extensions): (extensions,)
    else:                          extensions = tuple(extensions)
    def _importer(f):
        global importers
        importers = importers.set(name, (f, extensions, sniff, pyr.m(auto=auto)))
        setattr(load, name, f)
        return f
    return _importer
def forget_importer(name):
    '''
    forget_importer(name) yields True if an importer of type name was successfully forgotten from
      the neuropythy importers list and false otherwise. This function must be called before an
      importer can be replaced.
    '''
    global importers
    name = name.lower()
    if name in importers:
        importers = importers.discard(name)
        delattr(load, name)
        return True
    else:
        return False



# The list of exporter types we understand
exporters = pyr.m()
'''
neuropythy.io.core.exporters is the persistent map of file-types that can be exported by neuropythy.
See also neuropythy.io.save.
'''

def guess_export_format(filename, data, **kwargs):
    '''
    guess_export_format(filename, data) attempts to guess the export file format for the given
      filename and data (to be exported); it does this guessing by looking at the file extension and
      using registered sniff-tests from exporters.  It will not attempt to save the file, so if the
      extension of the filename is missing, it is less likely that this function will deduce the
      file-type (though save will often succeeed at extracting the data by trying all types
      exhaustively). If guess_export_format cannot deduce the format, it yields None.

    Note that if the filename has an extension that is recognized by neuropythy but the data itself
    is inappropriate for that format, this function will never look beyond the extention in the
    filename; neither this function nor save perform that level of deduction.
    
    Keyword arguments that are passed to save should also be passed to guess_export_format.
    '''
    
    # First try file endings
    (_,filename) = os.path.split(filename)
    fnm = filename.lower()
    fmt = None
    # to make sure we get the most specific ending, sort the exporters by their length
    es = sorted(((k,e) for (k,es) in six.iteritems(exporters) for e in es[1]),
                key=lambda x:-len(x[1]))
    for (k,e) in es:
        if fnm.endswith(('.' + e) if e[0] != '.' else e):
            return k
    # that didn't work; let's check the sniffers
    for (k,(_,_,sniff)) in six.iteritems(exporters):
        try:
            if sniff(filename, data, **kwargs): return k
        except Exception: pass
    return None
def save(filename, data, format=None, **kwargs):
    '''
    save(filename, data) writes the given data to the given filename then yieds that filename.
    save(filename, data, format) specifies that the given format should be used; this should be the
      name of the exporter (though a file extension that is recognized also will work).

    Additionally, functions located in save.<format> may be used; so, for example, the following
    are equivalent calls:
      save(filename, image, 'nifti')
      save.nifti(filename, image)
    In fact, the save.nifti function is just the nifti exporter, so help(save.nifti) will also
    yield documentation for the nifti exporter.

    Keyword options may be passed to save; these must match those accepted by the given export
    function.
    '''
    filename = os.path.expanduser(os.path.expandvars(filename))
    if format is None:
        format = guess_export_format(filename, data, **kwargs)
        if format is None:
            raise ValueError('Could not deduce export format for file %s' % filename)
    else:
        format = format.lower()
        if format not in exporters:
            # it might be an extension
            fmt = next((k for (k,(_,es,_)) in six.iteritems(exporters) if format in es), None)
            if fmt is None:
                # okay, no idea what it is
                raise ValueError('Format \'%s\' not recognized by neuropythy' % format)
            format = fmt
    (f,_,_) = exporters[format]
    return f(filename, data, **kwargs)
def exporter(name, extensions=None, sniff=None):
    '''
    @exporter(name) is a decorator that declares that the following function is an file saveing
      function that should be registered with the neuropythy save function. See also the
      forget_exporter function.

    Any exporter function must take, as its first argument, a filename and, as its second argument,
    the object to be exported; after that it may take any number of keyword arguments, but no other
    non-keyword arguments. These keyword arguments can be passed to the neuropythy save function.
    
    The following options are accepted:
      * extensions (default: None) may be a string or a collection of strings that indicate possible
        file extensions for files of this type.
      * sniff (default: None) may optionally be a function f(s, d) that yields True when the given
        string s is a filename for a file of this type and/or the given object d is an object that
        can be exported as this type. If no sniff is given, this type can still be detected by
        running all exporters exhaustively and catching any raised exceptions; though this may
        result in partial files written to disk, so is not used by save.
    '''
    name = name.lower()
    if name in exporters:
        raise ValueError('An exporter for type %s already exists; use forget_exporter' % name)
    extensions = (extensions,) if pimms.is_str(extensions) else \
                 ()            if extensions is None       else \
                 tuple(extensions)
    def _exporter(f):
        global exporters
        exporters = exporters.set(name, (f, extensions, sniff))
        setattr(save, name, f)
        return f
    return _exporter
def forget_exporter(name):
    '''
    forget_exporter(name) yields True if an exporter of type name was successfully forgotten from
      the neuropythy exporters list and false otherwise. This function must be called before an
      exporter can be replaced.
    '''
    global exporters
    name = name.lower()
    if name in exporters:
        exporters = exporters.discard(name)
        delattr(save, name)
        return True
    else:
        return False

####################################################################################################
# General/universal importers/exporters

# JSON: used with neuropythy.util's normalize/denormalize system
@importer('json', ('json', 'json.gz', 'json.bz2', 'json.lzma'))
def load_json(filename, to='auto'):
    '''
    load_json(filename) yields the object represented by the json file or stream object filename.
    
    The optional argument to may be set to None to indicate that the JSON data should be returned
    verbatim rather than parsed by neuropythy's denormalize system.
    '''
    from neuropythy.util import denormalize as denorm
    if pimms.is_str(filename):
        try:
            with gzip.open(filename, 'rt') as fl: dat = json.load(fl)
        except Exception:
            with open(filename, 'rt') as fl: dat = json.load(fl)
    else:
        dat = json.load(filename)
        filename = '<stream>'
    if to is None: return dat
    elif to == 'auto': return denorm(dat)
    else: raise ValueError('unrecognized to option: %s' % to)
@exporter('json', ('json', 'json.gz', 'json.bz2', 'json.lzma'))
def save_json(filename, obj, normalize=True):
    '''
    save_json(filename, obj) writes the given object to the given filename (or stream) in a
      normalized JSON format.

    The optional argument normalize (default True) may be set to False to prevent the object from
    being run through neuropythy's normalize system.
    '''
    from neuropythy.util import normalize as norm
    dat = norm(obj) if normalize else obj
    if pimms.is_str(filename):
        jsonstr = json.dumps(dat)
        if any(filename.endswith(s) for s in ('.gz', '.bz2', '.lzma')):
            with gzip.open(filename, 'wt') as fl: fl.write(jsonstr)
        else:
            with open(filename, 'wt') as fl: fl.write(jsonstr)
    else: json.dump(dat, filename)
    return filename
@importer('csv', ('csv', 'csv.gz', 'csv.bz2', 'csv.lzma'))
def load_csv(filename, **kw):
    '''
    load_csv(filename) yields a pandas dataframe of the contents of the CSV file. If pandas cannot
      be loaded, then an error is raised.

    All optional arguments are passed along to the pandas.read_csv function.
    '''
    import pandas
    if any(filename.endswith(s) for s in ('.gz', '.bz2', '.lzma')):
        with gzip.open(filename, 'rt') as fl: data = pandas.read_csv(fl, **kw)
    else:
        with open(filename, 'rt') as fl: data = pandas.read_csv(fl, **kw)
    return data
@exporter('csv', ('csv', 'csv.gz', 'csv.bz2', 'csv.lzma'))
def save_csv(filename, dat, index=False, **kw):
    '''
    save_csv(filename, d) writes a pandas dataframe d to a CSV file with the given name. If pandas
      cannot be loaded, then an error is raised. If d is not a dataframe, to_dataframe() is called
      on it.

    All optional arguments are passed along to the pandas.DataFrame.to_csv function.
    '''
    import pandas
    from neuropythy.util import to_dataframe
    d = to_dataframe(dat)
    if any(filename.endswith(s) for s in ('.gz', '.bz2', '.lzma')):
        with gzip.open(filename, 'wt', newlines='') as fl: d.to_csv(fl, index=index, **kw)
    else:
        with open(filename, 'wt') as fl: d.to_csv(fl, index=index, **kw)
    return filename
@importer('tsv', ('tsv', 'tsv.gz', 'tsv.bz2', 'tsv.lzma'))
def load_tsv(filename, sep='\t', **kw):
    '''
    load_tsv(filename) yields a pandas dataframe of the contents of the TSV file. If pandas cannot
      be loaded, then an error is raised.

    All optional arguments are passed along to the pandas.read_csv function. Note that this function
    is identical to the load_csv() function except that it has a default sep value of '\t' instead
    of ','.
    '''
    import pandas
    if any(filename.endswith(s) for s in ('.gz', '.bz2', '.lzma')):
        with gzip.open(filename, 'rt') as fl: data = pandas.read_csv(fl, sep=sep, **kw)
    else:
        with open(filename, 'rt') as fl: data = pandas.read_csv(fl, sep=sep, **kw)
    return data
@exporter('tsv', ('tsv', 'tsv.gz', 'tsv.bz2', 'tsv.lzma'))
def save_tsv(filename, dat, sep='\t', index=False, **kw):
    '''
    save_tsv(filename, d) writes a pandas dataframe d to a TSV file with the given name. If pandas
      cannot be loaded, then an error is raised. If d is not a dataframe, to_dataframe() is called
      on it.

    All optional arguments are passed along to the pandas.DataFrame.to_csv function. Note that this
    function is identical to save_csv() except that it has a default sep value of '\t' instead of
    ','.
    '''
    import pandas
    from neuropythy.util import to_dataframe
    d = to_dataframe(dat)
    if any(filename.endswith(s) for s in ('.gz', '.bz2', '.lzma')):
        with gzip.open(filename, 'wt', newlines='') as fl: d.to_csv(fl, sep=sep, index=index, **kw)
    else:
        with open(filename, 'wt') as fl: d.to_csv(fl, sep=sep, index=index, **kw)
    return filename

# Nifti!
@importer('nifti', ('nii', 'nii.gz', 'nii.bz2', 'nii.lzma'))
def load_nifti(filename, to='auto'):
    '''
    load_nifti(filename) yields the Nifti1Image or Nifti2Image referened by the given filename by
      using the nibabel load function.
    
    The optional argument to may be used to coerce the resulting data to a particular format; the
    following arguments are understood:
      * 'header' will yield just the image header
      * 'data' will yield the image's data-array
      * 'field' will yield a squeezed version of the image's data-array and will raise an error if
        the data object has more than 2 non-unitary dimensions (appropriate for loading surface
        properties stored in image files)
      * 'affine' will yield the image's affine transformation
      * 'image' will yield the raw image object
      * 'auto' is equivalent to 'image' unless the image has no more than 2 non-unitary dimensions,
        in which case it is assumed to be a surface-field and the return value is equivalent to
        the 'field' value.
    '''
    img = nib.load(filename)
    to = to.lower()
    if to == 'image':    return img
    elif to == 'data':   return img.dataobj
    elif to == 'affine': return img.affine
    elif to == 'header': return img.header
    elif to == 'field':
        dat = np.squeeze(np.asarray(img.dataobj))
        if len(dat.shape) > 2:
            raise ValueError('image requested as field has more than 2 non-unitary dimensions')
        return dat
    elif to in ['auto', 'automatic']:
        dims = set(np.shape(img.dataobj))
        if 1 < len(dims) < 4 and 1 in dims:
            return np.squeeze(np.asarray(img.dataobj))
        else:
            return img
    else:
        raise ValueError('unrecognized \'to\' argument \'%s\'' % to)
def to_nifti(obj, like=None, header=None, affine=None, extensions=Ellipsis, version=None):
    '''
    to_nifti(obj) yields a Nifti2Image object that is as equivalent as possible to the given object
      obj. If obj is a Nifti2Image already, then it is returned unmolested; other deduction rules
      are described below.

    The following options are accepted:
      * like (default: None) may be provided to give a guide for the various header- and meta-data
        that is included in the image. If this is a nifti image object, its meta-data are used; if
        this is a subject, then the meta-data are deduced from the subject's voxel and native
        orientation matrices. All other specific options below override anything deduced from the
        like argument.
      * header (default: None) may be a Nifti1 or Niti2 image header to be used as the nifti header
        or to replace the header in a new image.
      * affine (default: None) may specify the affine transform to be given to the image object.
      * extensions (default: Ellipsis) may specify a nifti extensions object that should be included
        in the header. The default value, Ellipsis, indicates that the extensions should not be
        changed, and that None should be used if extensions are not implied in obj (if, for example,
        obj is a data array rather than an image object with a header already.
      * version (default: None) may be specified as 1 or 2 for a Nifti1Image or Nifti2Image object,
        respectively; if the option None is passed, then any object that is already a nifti1 or
        nifti2 object is kept as the same version, otherwise nifti1 is used when possible and nifti2
        when not possible.
    '''
    from neuropythy.mri import (Subject, to_image)
    obj0 = obj
    # First go from like to explicit versions of affine and header:
    if like is not None:
        if isinstance(like, nib.analyze.AnalyzeHeader) or \
           isinstance(like, nib.freesurfer.mghformat.MGHHeader):
            if header is None: header = like
        elif isinstance(like, nib.analyze.SpatialImage):
            if header is None: header = like.header
            if affine is None: affine = like.affine
        elif isinstance(like, Subject):
            if affine is None: affine = like.images['brain'].affine
        else:
            raise ValueError('Could not interpret like argument with type %s' % type(like))
    # Figure out what the data is
    if isinstance(obj, nib.analyze.SpatialImage):
        obj = obj.dataobj
    elif not pimms.is_nparray(obj):
        obj = np.asarray(obj)
    if   len(obj.shape) == 3: obj = np.reshape(obj, obj.shape + (1,))
    elif len(obj.shape) == 2: obj = np.reshape(obj, obj.shape + (1,1))
    elif len(obj.shape) == 1: obj = np.reshape(obj, obj.shape + (1,1,1))
    elif len(obj.shape) != 4: raise ValueError('nifti objects must be 1D, 2D, 3D, or 4D')
    # figure out version if needed
    if version is None:
        if   isinstance(obj0, nib.nifti2.Nifti2Image): version = 2
        elif any(sh > 32767 for sh in obj.shape):      version = 2
        else:                                          version = 1
    # check to make sure that we have to change something:
    elif ((version == 1 and (isinstance(obj0, nib.nifti1.Nifti1Image) and
                             not isinstance(obj0, nib.nifti2.Nifti2Image))) or
          (version == 2 and isinstance(obj0, nib.nifti2.Nifti2Image))):
        if ((header is None or obj0.header is header) and
            (extensions is Ellipsis or extensions is obj0.header.extensions or
             (extensions is None and len(obj0.header.extensions) == 0))):
            return obj0
    # okay, now look at the header and affine etc.
    if header is None:
        if isinstance(obj0, nib.analyze.SpatialImage):
            header = obj0.header
        else:
            header = nib.nifti1.Nifti1Header() if version == 1 else nib.nifti2.Nifti2Header()
    if affine is None:
        if isinstance(obj0, nib.analyze.SpatialImage):
            affine = obj0.affine
        else:
            affine = np.eye(4)
    if extensions is None:
        extensions = nib.nifti1.Nifti1Extensions()
    # Okay, make a new object now...
    if version == 1:
        obj = nib.nifti1.Nifti1Image(obj, affine, header)
    elif version == 2:
        obj = nib.nifti2.Nifti2Image(obj, affine, header)
    else:
        raise ValueError('invalid version given (should be 1 or 2): %s' % version)
    # add the extensions if they're needed
    if extensions is not Ellipsis and (len(extensions) > 0 or len(obj.header.extensions) > 0):
        obj.header.extensions = extensions
    # Okay, that's it!
    return obj
@exporter('nifti', ('nii', 'nii.gz'))
def save_nifti(filename, obj, like=None, header=None, affine=None, extensions=Ellipsis,
               version=None):
    '''
    save_nifti(filename, obj) saves the given object to the given filename in the nifti format and
      returns the filename.

    All options that can be given to the to_nifti function can also be passed to this function; they
    are used to modify the object prior to exporting it.   
    '''
    obj = to_nifti(obj, like=like, header=header, affine=affine, version=version,
                   extensions=extensions)
    obj.to_filename(filename)
    return filename

@importer('string', ('txt','text'), auto=False)
def load_string(filename, to=None):
    '''
    load_string(filename) loads the given file as a string. The optional argument to can be set to
      'lines' to load a list of lines or to 'bytes' to load the data as a byte-string.
    '''
    if to is None: to = 'string'
    to = to.lower()
    if to in ['string', 'str', 'text', 'txt', 't', 's']:
        with open(filename, 'r') as fl:
            return fl.read()
    elif to in ['binary', 'bytes', 'b', 'raw']:
        with open(filename, 'rb') as fl:
            return fl.read()
    elif to in ['lines', 'l']:
        with open(filename, 'r') as fl:
            return fl.read().splitlines()
@exporter('string', ('txt','text'))
def save_string(filename, s):
    '''
    save_string(filename, s) saves the given string to the given file as text. The argument s may
      be a list of strings (lines). If s is neither a list of strings nor a string, it is cast to
      a string using str(s).
    '''
    if pimms.is_vector(s, 'string'):
        with open(filename, 'w') as fl:
            for l in s: fl.write(s + '\n')
        return filename
    elif not pimms.is_str(s): s = str(s)
    with open(filename, 'w') as fl:
        fl.write(s)
    return filename
@importer('bytes', ('bin', 'bytes'), auto=False)
def load_bytes(filename):
    '''
    load_bytes(filename) loads the given file as a byte-string.
    '''
    with open(filename, 'rb') as fl:
        return fl.read()
@exporter('bytes', ('bin','bytes'))
def save_bytes(filename, s):
    '''
    save_bytes(filename, s) saves the given byte-string s to the given file as raw bytes. The
      argument s must be either bytes or a bytearray.
    '''
    with open(filename, 'wb') as fl: fl.write(s)
    return filename

    



####################################################################################################
# neuropythy/io/__init__.py

'''
neuropythy.io is a namespace that contains tools for loading and saving data in neuroscientific
formats. It is intended as an extension of the nibabel libraries in that it is good at
auto-detecting many common formats and data-types and yields data in the neuropythy object system.
'''

from .core import (load, save, importer, exporter, forget_importer, forget_exporter,
                   to_nifti, load_json, save_json, load_csv, save_csv, load_tsv, save_tsv)

